import sys
from contextlib import contextmanager
from pathlib import Path
from stat import S_IWGRP, S_IWOTH, S_IWUSR
from textwrap import dedent
from typing import Callable, Iterator, NamedTuple

import pytest
from packaging.requirements import Requirement
from pytest_mock import MockerFixture

from tox.pytest import TempPathFactory
from tox.util.pep517.frontend import BackendFailed
from tox.util.pep517.via_fresh_subprocess import SubprocessFrontend

if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
    from importlib.metadata import Distribution, EntryPoint  # type: ignore[attr-defined]
else:  # pragma: no cover (<py38)
    from importlib_metadata import Distribution, EntryPoint  # noqa


@pytest.fixture(scope="session")
def frontend_setuptools(tmp_path_factory: TempPathFactory) -> SubprocessFrontend:
    prj = tmp_path_factory.mktemp("proj")
    (prj / "pyproject.toml").write_text(
        '[build-system]\nrequires=["setuptools","wheel"]\nbuild-backend = "setuptools.build_meta"'
    )
    cfg = """
        [metadata]
        name = demo
        version = 1.0

        [options]
        packages = demo
        install_requires =
          requests>2
          magic>3

        [options.entry_points]
        console_scripts =
            demo_exe = demo:a
        """
    (prj / "setup.cfg").write_text(cfg)
    (prj / "setup.py").write_text("from setuptools import setup; setup()")
    demo = prj / "demo"
    demo.mkdir()
    (demo / "__init__.py").write_text("def a(): print('ok')")
    args = SubprocessFrontend.create_args_from_folder(prj)
    return SubprocessFrontend(*args[:-1])


def test_pep517_setuptools_get_requires_for_build_sdist(frontend_setuptools: SubprocessFrontend) -> None:
    result = frontend_setuptools.get_requires_for_build_sdist()
    assert result.requires == ()
    assert isinstance(result.out, str)
    assert isinstance(result.err, str)


def test_pep517_setuptools_get_requires_for_build_wheel(frontend_setuptools: SubprocessFrontend) -> None:
    result = frontend_setuptools.get_requires_for_build_wheel()
    for left, right in zip(result.requires, (Requirement("wheel"),)):
        assert isinstance(left, Requirement)
        assert str(left) == str(right)
    assert isinstance(result.out, str)
    assert isinstance(result.err, str)


def test_pep517_setuptools_prepare_metadata_for_build_wheel(
    frontend_setuptools: SubprocessFrontend, tmp_path: Path
) -> None:
    meta = tmp_path / "meta"
    result = frontend_setuptools.prepare_metadata_for_build_wheel(metadata_directory=meta)
    dist = Distribution.at(str(result.metadata))
    assert dist.entry_points == [EntryPoint(name="demo_exe", value="demo:a", group="console_scripts")]
    assert dist.version == "1.0"
    assert dist.metadata["Name"] == "demo"
    assert [v for k, v in dist.metadata.items() if k == "Requires-Dist"] == ["requests (>2)", "magic (>3)"]
    assert isinstance(result.out, str)
    assert isinstance(result.err, str)

    # call it again regenerates it because frontend always deletes earlier content
    before = result.metadata.stat().st_mtime
    result = frontend_setuptools.prepare_metadata_for_build_wheel(metadata_directory=meta)
    after = result.metadata.stat().st_mtime
    assert after > before


def test_pep517_setuptools_build_sdist(frontend_setuptools: SubprocessFrontend, tmp_path: Path) -> None:
    result = frontend_setuptools.build_sdist(tmp_path)
    sdist = result.sdist
    assert sdist.exists()
    assert sdist.is_file()
    assert sdist.name == "demo-1.0.tar.gz"
    assert isinstance(result.out, str)
    assert isinstance(result.err, str)


def test_pep517_setuptools_build_wheel(frontend_setuptools: SubprocessFrontend, tmp_path: Path) -> None:
    result = frontend_setuptools.build_wheel(tmp_path)
    wheel = result.wheel
    assert wheel.exists()
    assert wheel.is_file()
    assert wheel.name == "demo-1.0-py3-none-any.whl"
    assert isinstance(result.out, str)
    assert isinstance(result.err, str)


def test_pep517_setuptools_exit(frontend_setuptools: SubprocessFrontend) -> None:
    result, out, err = frontend_setuptools.send_cmd("_exit")
    assert isinstance(out, str)
    assert isinstance(err, str)
    assert result == 0


def test_pep517_setuptools_missing_command(frontend_setuptools: SubprocessFrontend) -> None:
    with pytest.raises(BackendFailed):
        frontend_setuptools.send_cmd("missing_command")


def test_pep517_setuptools_exception(frontend_setuptools: SubprocessFrontend) -> None:
    with pytest.raises(BackendFailed) as context:
        frontend_setuptools.send_cmd("build_wheel")
    assert isinstance(context.value.out, str)
    assert isinstance(context.value.err, str)
    assert context.value.exc_type == "TypeError"
    assert context.value.exc_msg == "build_wheel() missing 1 required positional argument: 'wheel_directory'"
    assert context.value.code == 1
    assert context.value.args == ()
    assert repr(context.value)
    assert str(context.value)
    assert repr(context.value) != str(context.value)


def test_pep517_bad_message(frontend_setuptools: SubprocessFrontend, tmp_path: Path) -> None:
    with frontend_setuptools._send_msg("bad_cmd", tmp_path / "a", "{") as status:
        while not status.done:
            pass
    out, err = status.out_err()
    assert out
    assert "Backend: incorrect request to backend: {" in err


def test_pep517_result_missing(frontend_setuptools: SubprocessFrontend, tmp_path: Path, mocker: MockerFixture) -> None:
    class _Result(NamedTuple):
        name: str

    @contextmanager
    def named_temporary_file(prefix: str) -> Iterator[_Result]:
        write = S_IWUSR | S_IWGRP | S_IWOTH
        base = tmp_path / prefix
        result = base.with_suffix(".json")
        result.write_text("")
        result.chmod(result.stat().st_mode & ~write)  # force json write to fail due to R/O
        patch = mocker.patch("tox.util.pep517.frontend.Path.exists", return_value=False)  # make it missing
        try:
            yield _Result(str(base))
        finally:
            patch.stop()
            result.chmod(result.stat().st_mode | write)  # cleanup
            result.unlink()

    mocker.patch("tox.util.pep517.frontend.NamedTemporaryFile", named_temporary_file)
    with pytest.raises(BackendFailed) as context:
        frontend_setuptools.send_cmd("_exit")
    exc = context.value
    assert exc.exc_msg == f"Backend response file {tmp_path / 'pep517__exit-.json'} is missing"
    assert exc.exc_type == "RuntimeError"
    assert exc.code == 1
    assert "Traceback" in exc.err
    assert "PermissionError" in exc.err


@pytest.fixture
def local_builder(tmp_path: Path) -> Callable[[str], Path]:
    def _f(content: str) -> Path:
        toml = '[build-system]\nrequires=[]\nbuild-backend = "build_tester"\nbackend-path=["."]'
        (tmp_path / "pyproject.toml").write_text(toml)
        (tmp_path / "build_tester.py").write_text(dedent(content))
        return tmp_path

    return _f


def test_pep517_missing_backend(local_builder: Callable[[str], Path]) -> None:
    tmp_path = local_builder("")
    toml = tmp_path / "pyproject.toml"
    toml.write_text('[build-system]\nrequires=[]\nbuild-backend = "build_tester"')
    fronted = SubprocessFrontend(*SubprocessFrontend.create_args_from_folder(tmp_path)[:-1])
    with pytest.raises(BackendFailed) as context:
        fronted.build_wheel(tmp_path / "wheel")
    exc = context.value
    assert exc.exc_type == "RuntimeError"
    assert exc.code == 1
    assert "failed to start backend" in exc.err
    assert "ModuleNotFoundError: No module named " in exc.err


@pytest.mark.parametrize("cmd", ["build_wheel", "build_sdist"])
def test_pep517_missing_required_cmd(cmd: str, local_builder: Callable[[str], Path]) -> None:
    tmp_path = local_builder("")
    fronted = SubprocessFrontend(*SubprocessFrontend.create_args_from_folder(tmp_path)[:-1])

    with pytest.raises(BackendFailed) as context:
        getattr(fronted, cmd)(tmp_path)
    exc = context.value
    assert f"has no attribute '{cmd}'" in exc.exc_msg
    assert exc.exc_type == "MissingCommand"


def test_pep517_empty_pyproject(tmp_path: Path) -> None:
    (tmp_path / "pyproject.toml").write_text("[build-system]")
    root, backend_paths, backend_module, backend_obj, requires, _ = SubprocessFrontend.create_args_from_folder(tmp_path)
    assert root == tmp_path
    assert backend_paths == ()
    assert backend_module == "setuptools.build_meta"
    assert backend_obj == "__legacy__"
    for left, right in zip(requires, (Requirement("setuptools>=40.8.0"), Requirement("wheel"))):
        assert isinstance(left, Requirement)
        assert str(left) == str(right)


def test_pep517_backend_no_prepare_wheel(tmp_path: Path, demo_pkg_inline: Path) -> None:
    fronted = SubprocessFrontend(*SubprocessFrontend.create_args_from_folder(demo_pkg_inline)[:-1])
    result = fronted.prepare_metadata_for_build_wheel(tmp_path)
    assert result.metadata.name == "demo_pkg_inline-1.0.0.dist-info"


def test_pep517_backend_build_sdist_demo_pkg_inline(tmp_path: Path, demo_pkg_inline: Path) -> None:
    fronted = SubprocessFrontend(*SubprocessFrontend.create_args_from_folder(demo_pkg_inline)[:-1])
    result = fronted.build_sdist(sdist_directory=tmp_path)
    assert result.sdist == tmp_path / "demo_pkg_inline-1.0.0.tar.gz"


def test_pep517_backend_obj(tmp_path: Path) -> None:
    (tmp_path / "pyproject.toml").write_text(
        dedent(
            """
        [build-system]
        requires=[]
        build-backend = "build.api:backend:"
        backend-path=["."]
        """
        )
    )
    build = tmp_path / "build"
    build.mkdir()
    (build / "__init__.py").write_text("")
    (build / "api.py").write_text(
        dedent(
            """
        class A:
            def get_requires_for_build_sdist(self, config_settings=None):
                return ["a"]

        backend = A()
        """
        )
    )
    fronted = SubprocessFrontend(*SubprocessFrontend.create_args_from_folder(tmp_path)[:-1])
    result = fronted.get_requires_for_build_sdist()
    for left, right in zip(result.requires, (Requirement("a"),)):
        assert isinstance(left, Requirement)
        assert str(left) == str(right)


@pytest.mark.parametrize("of_type", ["wheel", "sdist"])
def test_pep517_get_requires_for_build_missing(of_type: str, local_builder: Callable[[str], Path]) -> None:
    tmp_path = local_builder("")
    fronted = SubprocessFrontend(*SubprocessFrontend.create_args_from_folder(tmp_path)[:-1])
    result = getattr(fronted, f"get_requires_for_build_{of_type}")()
    assert result.requires == ()


@pytest.mark.parametrize("of_type", ["sdist", "wheel"])
def test_pep517_bad_return_type_get_requires_for_build(of_type: str, local_builder: Callable[[str], Path]) -> None:
    tmp_path = local_builder(f"def get_requires_for_build_{of_type}(config_settings=None): return 1")
    fronted = SubprocessFrontend(*SubprocessFrontend.create_args_from_folder(tmp_path)[:-1])

    with pytest.raises(BackendFailed) as context:
        getattr(fronted, f"get_requires_for_build_{of_type}")()

    exc = context.value
    msg = f"'get_requires_for_build_{of_type}' on 'build_tester' returned 1 but expected type 'list of string'"
    assert exc.exc_msg == msg
    assert exc.exc_type == "TypeError"


def test_pep517_bad_return_type_build_sdist(local_builder: Callable[[str], Path]) -> None:
    tmp_path = local_builder("def build_sdist(sdist_directory, config_settings=None): return 1")
    fronted = SubprocessFrontend(*SubprocessFrontend.create_args_from_folder(tmp_path)[:-1])

    with pytest.raises(BackendFailed) as context:
        fronted.build_sdist(tmp_path)

    exc = context.value
    assert exc.exc_msg == f"'build_sdist' on 'build_tester' returned 1 but expected type {str!r}"
    assert exc.exc_type == "TypeError"


def test_pep517_bad_return_type_build_wheel(local_builder: Callable[[str], Path]) -> None:
    txt = "def build_wheel(wheel_directory, config_settings=None, metadata_directory=None): return 1"
    tmp_path = local_builder(txt)
    fronted = SubprocessFrontend(*SubprocessFrontend.create_args_from_folder(tmp_path)[:-1])

    with pytest.raises(BackendFailed) as context:
        fronted.build_wheel(tmp_path)

    exc = context.value
    assert exc.exc_msg == f"'build_wheel' on 'build_tester' returned 1 but expected type {str!r}"
    assert exc.exc_type == "TypeError"


def test_pep517_bad_return_type_prepare_metadata_for_build_wheel(local_builder: Callable[[str], Path]) -> None:
    tmp_path = local_builder("def prepare_metadata_for_build_wheel(metadata_directory, config_settings=None): return 1")
    fronted = SubprocessFrontend(*SubprocessFrontend.create_args_from_folder(tmp_path)[:-1])

    with pytest.raises(BackendFailed) as context:
        fronted.prepare_metadata_for_build_wheel(tmp_path / "meta")

    exc = context.value
    assert exc.exc_type == "TypeError"
    assert exc.exc_msg == f"'prepare_metadata_for_build_wheel' on 'build_tester' returned 1 but expected type {str!r}"


def test_pep517_prepare_metadata_for_build_wheel_meta_is_root(local_builder: Callable[[str], Path]) -> None:
    tmp_path = local_builder("def prepare_metadata_for_build_wheel(metadata_directory, config_settings=None): return 1")
    fronted = SubprocessFrontend(*SubprocessFrontend.create_args_from_folder(tmp_path)[:-1])

    with pytest.raises(RuntimeError) as context:
        fronted.prepare_metadata_for_build_wheel(tmp_path)

    assert str(context.value) == f"the project root and the metadata directory can't be the same {tmp_path}"


def test_pep517_no_wheel_prepare_metadata_for_build_wheel(local_builder: Callable[[str], Path]) -> None:
    txt = "def build_wheel(wheel_directory, config_settings=None, metadata_directory=None): return 'out'"
    tmp_path = local_builder(txt)
    fronted = SubprocessFrontend(*SubprocessFrontend.create_args_from_folder(tmp_path)[:-1])

    with pytest.raises(RuntimeError, match="missing wheel file return by backed *"):
        fronted.prepare_metadata_for_build_wheel(tmp_path / "meta")


def test_pep517_bad_wheel_prepare_metadata_for_build_wheel(local_builder: Callable[[str], Path]) -> None:
    txt = """
    import sys
    from pathlib import Path
    from zipfile import ZipFile

    def build_wheel(wheel_directory, config_settings=None, metadata_directory=None):
        path = Path(wheel_directory) / "out"
        with ZipFile(str(path), "w") as zip_file_handler:
            pass
        print(f"created wheel {path}")
        return path.name
    """
    tmp_path = local_builder(txt)
    fronted = SubprocessFrontend(*SubprocessFrontend.create_args_from_folder(tmp_path)[:-1])

    with pytest.raises(RuntimeError, match="no .dist-info found inside generated wheel*"):
        fronted.prepare_metadata_for_build_wheel(tmp_path / "meta")



from tox.config.main import Config
from tox.config.of_type import ConfigConstantDefinition, ConfigDynamicDefinition


def test_config_constant_eq() -> None:
    val_1 = ConfigConstantDefinition(("key",), "description", "env", "value")
    val_2 = ConfigConstantDefinition(("key",), "description", "env", "value")
    assert val_1 == val_2


def test_config_dynamic_eq() -> None:
    def func(name: str, _: Config) -> str:
        return name  # pragma: no cover

    val_1 = ConfigDynamicDefinition(("key",), "description", "env", str, "default", post_process=func)
    val_2 = ConfigDynamicDefinition(("key",), "description", "env", str, "default", post_process=func)
    assert val_1 == val_2



"""Helper methods related to the CPU"""
import multiprocessing
from typing import Optional


def auto_detect_cpus() -> int:
    try:
        n: Optional[int] = multiprocessing.cpu_count()
    except NotImplementedError:
        n = None
    return n if n else 1


__all__ = ("auto_detect_cpus",)



"""
Declare and handle the tox env info file (a file at the root of every tox environment that contains information about
the status of the tox environment - python version of the environment, installed packages, etc.).
"""
import json
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Iterator, Optional, Tuple


class Info:
    def __init__(self, path: Path) -> None:
        self._path = path / ".tox-info.json"
        try:
            value = json.loads(self._path.read_text())
        except (ValueError, OSError):
            value = {}
        self._content = value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(path={self._path})"

    @contextmanager
    def compare(
        self, value: Any, section: str, sub_section: Optional[str] = None
    ) -> Iterator[Tuple[bool, Optional[Any]]]:
        """Cache"""
        old = self._content.get(section)
        if sub_section is not None and old is not None:
            old = old.get(sub_section)

        if old == value:
            yield True, None
        else:
            yield False, old
            # if no exception thrown update
            if sub_section is None:
                self._content[section] = value
            else:
                if self._content.get(section) is None:
                    self._content[section] = {sub_section: value}
                else:
                    self._content[section][sub_section] = value
            self._write()

    def reset(self) -> None:
        self._content = {}

    def _write(self) -> None:
        self._path.write_text(json.dumps(self._content, indent=2))


__all__ = ("Info",)



from collections import OrderedDict
from typing import Any, Iterator, List, Sequence

from tox.execute.request import shell_cmd


class Command:
    def __init__(self, args: List[str]) -> None:
        self.ignore_exit_code = args[0] == "-"
        self.args = args[1:] if self.ignore_exit_code else args

    def __repr__(self) -> str:
        return f"{type(self).__name__}(args={(['-'] if self.ignore_exit_code else [])+ self.args!r})"

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and (self.args, self.ignore_exit_code) == (other.args, other.ignore_exit_code)

    def __ne__(self, other: Any) -> bool:
        return not (self == other)

    @property
    def shell(self) -> str:
        return shell_cmd(self.args)


class EnvList:
    def __init__(self, envs: Sequence[str]) -> None:
        self.envs = list(OrderedDict((e, None) for e in envs).keys())

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.envs!r})"

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.envs == other.envs

    def __ne__(self, other: Any) -> bool:
        return not (self == other)

    def __iter__(self) -> Iterator[str]:
        return iter(self.envs)

    def __bool__(self) -> bool:
        return bool(self.envs)


__all__ = (
    "Command",
    "EnvList",
)



"""
Defines how to run a single tox environment.
"""
import logging
import time
from typing import List, NamedTuple, Tuple, cast

from tox.config.types import Command
from tox.execute.api import Outcome, StdinSource
from tox.tox_env.api import ToxEnv
from tox.tox_env.errors import Fail, Skip
from tox.tox_env.python.virtual_env.package.api import ToxBackendFailed
from tox.tox_env.runner import RunToxEnv

LOGGER = logging.getLogger(__name__)


class ToxEnvRunResult(NamedTuple):
    name: str
    skipped: bool
    code: int
    outcomes: List[Outcome]
    duration: float


def run_one(tox_env: RunToxEnv, recreate: bool, no_test: bool, suspend_display: bool) -> ToxEnvRunResult:
    start_one = time.monotonic()
    name = cast(str, tox_env.conf.name)
    with tox_env.display_context(suspend_display):
        skipped, code, outcomes = _evaluate(tox_env, recreate, no_test)
    duration = time.monotonic() - start_one
    return ToxEnvRunResult(name, skipped, code, outcomes, duration)


def _evaluate(tox_env: ToxEnv, recreate: bool, no_test: bool) -> Tuple[bool, int, List[Outcome]]:
    skipped = False
    code: int = 0
    outcomes: List[Outcome] = []
    try:
        try:
            tox_env.ensure_setup(recreate=recreate)
            code, outcomes = run_commands(tox_env, no_test)
        except Skip as exception:
            LOGGER.info("skipped environment because %s", exception)
            skipped = True
        except ToxBackendFailed as exception:
            LOGGER.error("%s", exception)
            raise SystemExit(exception.code)
        except Fail as exception:
            LOGGER.error("failed with %s", exception)
            code = 1
        except Exception:  # noqa # pragma: no cover
            LOGGER.exception("internal error")  # pragma: no cover
            code = 2  # pragma: no cover
        finally:
            tox_env.teardown()
    except SystemExit as exception:  # setup command fails (interrupted or via invocation)
        code = exception.code
    return skipped, code, outcomes


def run_commands(tox_env: ToxEnv, no_test: bool) -> Tuple[int, List[Outcome]]:
    status = Outcome.OK  # assume all good
    outcomes: List[Outcome] = []
    if no_test is False:
        keys = ("commands_pre", "commands", "commands_post")
        for key in keys:
            for at, cmd in enumerate(cast(List[Command], tox_env.conf[key])):
                current_outcome = tox_env.execute(
                    cmd.args,
                    cwd=tox_env.conf["change_dir"],
                    stdin=StdinSource.user_only(),
                    show=True,
                    run_id=f"{key}[{at}]",
                )
                outcomes.append(current_outcome)
                if cmd.ignore_exit_code:
                    continue
                try:
                    current_outcome.assert_success()
                except SystemExit as exception:
                    return exception.code, outcomes
    return status, outcomes


__all__ = (
    "run_one",
    "ToxEnvRunResult",
)



from configparser import ConfigParser
from typing import Callable

import pytest

from tox.config.loader.api import Override
from tox.config.loader.ini import IniLoader


def test_ini_loader_keys(mk_ini_conf: Callable[[str], ConfigParser]) -> None:
    loader = IniLoader("tox", mk_ini_conf("\n[tox]\n\na=b\nc=d\n\n"), [])
    assert loader.found_keys() == {"a", "c"}


def test_ini_loader_repr(mk_ini_conf: Callable[[str], ConfigParser]) -> None:
    loader = IniLoader("tox", mk_ini_conf("\n[tox]\n\na=b\nc=d\n\n"), [Override("tox.a=1")])
    assert repr(loader) == "IniLoader(section=<Section: tox>, overrides={'a': Override('tox.a=1')})"


def test_ini_loader_has_section(mk_ini_conf: Callable[[str], ConfigParser]) -> None:
    loader = IniLoader("tox", mk_ini_conf("[magic]\n[tox]\n\na=b\nc=d\n\n"), [])
    assert loader.get_section("magic") is not None


def test_ini_loader_has_no_section(mk_ini_conf: Callable[[str], ConfigParser]) -> None:
    loader = IniLoader("tox", mk_ini_conf("[tox]\n\na=b\nc=d\n\n"), [])
    assert loader.get_section("magic") is None


def test_ini_loader_raw(mk_ini_conf: Callable[[str], ConfigParser]) -> None:
    loader = IniLoader("tox", mk_ini_conf("[tox]\na=b"), [])
    result = loader.load(key="a", of_type=str, conf=None, env_name=None)
    assert result == "b"


@pytest.mark.parametrize("sep", ["\n", "\r\n"])
def test_ini_loader_raw_strip_escaped_newline(mk_ini_conf: Callable[[str], ConfigParser], sep: str) -> None:
    loader = IniLoader("tox", mk_ini_conf(f"[tox]{sep}a=b\\{sep} c"), [])
    result = loader.load(key="a", of_type=str, conf=None, env_name=None)
    assert result == "bc"



"""Module declaring a command execution request."""
import sys
from enum import Enum
from pathlib import Path
from typing import Dict, List, Sequence, Union


class StdinSource(Enum):
    OFF = 0
    USER = 1
    API = 2

    @staticmethod
    def user_only() -> "StdinSource":
        return StdinSource.USER if sys.stdin.isatty() else StdinSource.OFF


class ExecuteRequest:
    """Defines a commands execution request"""

    def __init__(
        self, cmd: Sequence[Union[str, Path]], cwd: Path, env: Dict[str, str], stdin: StdinSource, run_id: str
    ) -> None:
        if len(cmd) == 0:
            raise ValueError("cannot execute an empty command")
        self.cmd: List[str] = [str(i) for i in cmd]
        self.cwd = cwd
        self.env = env
        self.stdin = stdin
        self.run_id = run_id

    @property
    def shell_cmd(self) -> str:
        try:
            exe = str(Path(self.cmd[0]).relative_to(self.cwd))
        except ValueError:
            exe = self.cmd[0]
        _cmd = [exe]
        _cmd.extend(self.cmd[1:])
        return shell_cmd(_cmd)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(cmd={self.cmd!r}, cwd={self.cwd!r}, env=..., stdin={self.stdin!r})"


def shell_cmd(cmd: Sequence[str]) -> str:
    if sys.platform == "win32":  # pragma: win32 cover
        from subprocess import list2cmdline

        return list2cmdline(tuple(str(x) for x in cmd))
    else:  # pragma: win32 no cover
        from shlex import quote as shlex_quote

        return " ".join(shlex_quote(str(x)) for x in cmd)


__all__ = (
    "StdinSource",
    "ExecuteRequest",
    "shell_cmd",
)



"""Execute that runs on local file system via subprocess-es"""
import logging
import os
import shutil
import sys
import time
from subprocess import DEVNULL, PIPE, TimeoutExpired
from types import TracebackType
from typing import TYPE_CHECKING, Generator, List, Optional, Sequence, Tuple, Type

from ..api import Execute, ExecuteInstance, ExecuteStatus
from ..request import ExecuteRequest, StdinSource
from ..stream import SyncWrite
from .read_via_thread import WAIT_GENERAL

if sys.platform == "win32":  # explicit check for mypy # pragma: win32 cover
    # needs stdin/stdout handlers backed by overlapped IO
    if TYPE_CHECKING:  # the typeshed libraries don't contain this, so replace it with normal one
        from subprocess import Popen
    else:
        from asyncio.windows_utils import Popen
    from signal import CTRL_C_EVENT as SIG_INTERRUPT
    from subprocess import CREATE_NEW_PROCESS_GROUP

    from .read_via_thread_windows import ReadViaThreadWindows as ReadViaThread

    CREATION_FLAGS = CREATE_NEW_PROCESS_GROUP  # a custom flag needed for Windows signal send ability (CTRL+C)
else:  # pragma: win32 no cover
    from signal import SIGINT as SIG_INTERRUPT
    from signal import SIGKILL, SIGTERM
    from subprocess import Popen

    from .read_via_thread_unix import ReadViaThreadUnix as ReadViaThread

    CREATION_FLAGS = 0

WAIT_INTERRUPT = 0.3
WAIT_TERMINATE = 0.2
IS_WIN = sys.platform == "win32"


class LocalSubProcessExecutor(Execute):
    def build_instance(self, request: ExecuteRequest, out: SyncWrite, err: SyncWrite) -> ExecuteInstance:
        return LocalSubProcessExecuteInstance(request, out, err)


class LocalSubprocessExecuteStatus(ExecuteStatus):
    def __init__(self, out: SyncWrite, err: SyncWrite, process: "Popen[bytes]"):
        self._process: "Popen[bytes]" = process
        super().__init__(out, err)
        self._interrupted = False

    @property
    def exit_code(self) -> Optional[int]:
        return self._process.returncode

    def interrupt(self) -> None:
        self._interrupted = True
        if self._process is not None:  # pragma: no branch
            # A three level stop mechanism for children - INT -> TERM -> KILL
            # communicate will wait for the app to stop, and then drain the standard streams and close them
            proc = self._process
            host_pid = os.getpid()
            to_pid = proc.pid
            logging.warning("requested interrupt of %d from %d", to_pid, host_pid)
            if proc.poll() is None:  # still alive, first INT
                logging.warning(
                    "send signal %s to %d from %d with timeout %.2f",
                    f"SIGINT({SIG_INTERRUPT})",
                    to_pid,
                    host_pid,
                    WAIT_INTERRUPT,
                )
                proc.send_signal(SIG_INTERRUPT)
                start = time.monotonic()
                while proc.poll() is None and (time.monotonic() - start) < WAIT_INTERRUPT:
                    continue
                if proc.poll() is None:  # pragma: no branch
                    if sys.platform == "win32":  # explicit check for mypy # pragma: no branch
                        logging.warning("terminate %d from %d", to_pid, host_pid)  # pragma: no cover
                    else:
                        logging.warning(
                            "send signal %s to %d from %d with timeout %.2f",
                            f"SIGTERM({SIGTERM})",
                            to_pid,
                            host_pid,
                            WAIT_TERMINATE,
                        )
                    proc.terminate()
                    start = time.monotonic()
                    if sys.platform != "win32":  # explicit check for mypy # pragma: no branch
                        # Windows terminate is UNIX kill
                        while proc.poll() is None and (time.monotonic() - start) < WAIT_TERMINATE:
                            continue
                        if proc.poll() is None:  # pragma: no branch
                            logging.warning("send signal %s to %d from %d", f"SIGKILL({SIGKILL})", to_pid, host_pid)
                            proc.kill()
                    while proc.poll() is None:
                        continue
            else:  # pragma: no cover # difficult to test, process must die just as it's being interrupted
                logging.warning("process already dead with %s within %s", proc.returncode, os.getpid())
            logging.warning("interrupt finished with success")

    def wait(self, timeout: Optional[float] = None) -> None:
        # note poll in general might deadlock if output large, but we drain in background threads so not an issue here
        try:
            self._process.wait(timeout=WAIT_GENERAL if timeout is None else timeout)
        except TimeoutExpired:
            pass

    def write_stdin(self, content: str) -> None:
        stdin = self._process.stdin
        if stdin is None:  # pragma: no branch
            return  # pragma: no cover
        bytes_content = content.encode()
        try:
            if sys.platform == "win32":  # explicit check for mypy  # pragma: win32 cover
                # on Windows we have a PipeHandle object here rather than a file stream
                import _overlapped  # type: ignore[import]

                ov = _overlapped.Overlapped(0)
                ov.WriteFile(stdin.handle, bytes_content)  # type: ignore[attr-defined]
                result = ov.getresult(10)  # wait up to 10ms to perform the operation
                if result != len(bytes_content):
                    raise RuntimeError(f"failed to write to {stdin!r}")
            else:
                stdin.write(bytes_content)
                stdin.flush()
        except OSError:  # pragma: no cover
            if self._interrupted:  # pragma: no cover
                pass  # pragma: no cover  # if the process was asked to exit in the meantime ignore write errors
            raise  # pragma: no cover

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(pid={self._process.pid}, returncode={self._process.returncode!r})"


class LocalSubprocessExecuteFailedStatus(ExecuteStatus):
    def __init__(self, out: SyncWrite, err: SyncWrite, exit_code: Optional[int]) -> None:
        super().__init__(out, err)
        self._exit_code = exit_code

    @property
    def exit_code(self) -> Optional[int]:
        return self._exit_code

    def wait(self, timeout: Optional[float] = None) -> None:
        """already dead no need to wait"""

    def write_stdin(self, content: str) -> None:
        """cannot write"""

    def interrupt(self) -> None:
        """Nothing running so nothing to interrupt"""


class LocalSubProcessExecuteInstance(ExecuteInstance):
    def __init__(
        self,
        request: ExecuteRequest,
        out: SyncWrite,
        err: SyncWrite,
        on_exit_drain: bool = True,
    ) -> None:
        super().__init__(request, out, err)
        self.process: Optional[Popen[bytes]] = None
        self._cmd: Optional[List[str]] = None
        self._read_stderr: Optional[ReadViaThread] = None
        self._read_stdout: Optional[ReadViaThread] = None
        self._on_exit_drain = on_exit_drain

    @property
    def cmd(self) -> Sequence[str]:
        if self._cmd is None:
            executable = shutil.which(self.request.cmd[0], path=self.request.env["PATH"])
            if executable is None:
                cmd = self.request.cmd  # if failed to find leave as it is
            else:
                # else use expanded format
                cmd = [executable, *self.request.cmd[1:]]
            self._cmd = cmd
        return self._cmd

    def __enter__(self) -> ExecuteStatus:
        # adjust sub-process terminal size
        columns, lines = shutil.get_terminal_size(fallback=(-1, -1))
        if columns != -1:  # pragma: no branch
            self.request.env["COLUMNS"] = str(columns)
        if columns != -1:  # pragma: no branch
            self.request.env["LINES"] = str(lines)

        stdout, stderr = self.get_stream_file_no("stdout"), self.get_stream_file_no("stderr")

        try:
            self.process = process = Popen(
                self.cmd,
                stdout=next(stdout),
                stderr=next(stderr),
                stdin={StdinSource.USER: None, StdinSource.OFF: DEVNULL, StdinSource.API: PIPE}[self.request.stdin],
                cwd=str(self.request.cwd),
                env=self.request.env,
                creationflags=CREATION_FLAGS,
            )
        except OSError as exception:
            return LocalSubprocessExecuteFailedStatus(self._out, self._err, exception.errno)

        status = LocalSubprocessExecuteStatus(self._out, self._err, process)
        drain, pid = self._on_exit_drain, self.process.pid
        self._read_stderr = ReadViaThread(stderr.send(process), self.err_handler, name=f"err-{pid}", drain=drain)
        self._read_stderr.__enter__()
        self._read_stdout = ReadViaThread(stdout.send(process), self.out_handler, name=f"out-{pid}", drain=drain)
        self._read_stdout.__enter__()

        if sys.platform == "win32":  # explicit check for mypy:  # pragma: win32 cover
            process.stderr.read = self._read_stderr._drain_stream  # type: ignore[assignment,union-attr]
            process.stdout.read = self._read_stdout._drain_stream  # type: ignore[assignment,union-attr]
        return status

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        if self._read_stderr is not None:
            self._read_stderr.__exit__(exc_type, exc_val, exc_tb)
        if self._read_stdout is not None:
            self._read_stdout.__exit__(exc_type, exc_val, exc_tb)

    @staticmethod
    def get_stream_file_no(key: str) -> Generator[int, "Popen[bytes]", None]:
        process = yield PIPE
        stream = getattr(process, key)
        if sys.platform == "win32":  # explicit check for mypy # pragma: win32 cover
            yield stream.handle
        else:
            yield stream.name

    def set_out_err(self, out: SyncWrite, err: SyncWrite) -> Tuple[SyncWrite, SyncWrite]:
        prev = self._out, self._err
        if self._read_stdout is not None:  # pragma: no branch
            self._read_stdout.handler = out.handler
        if self._read_stderr is not None:  # pragma: no branch
            self._read_stderr.handler = err.handler
        return prev


__all__ = (
    "SIG_INTERRUPT",
    "CREATION_FLAGS",
    "LocalSubProcessExecuteInstance",
    "LocalSubProcessExecutor",
    "LocalSubprocessExecuteStatus",
    "LocalSubprocessExecuteFailedStatus",
)



"""Build frontend for PEP-517"""
import json
import shutil
from abc import ABC, abstractmethod
from contextlib import contextmanager
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from time import sleep
from typing import Any, Dict, Iterator, List, NamedTuple, NoReturn, Optional, Tuple, cast
from zipfile import ZipFile

import toml
from packaging.requirements import Requirement

_HERE = Path(__file__).parent
ConfigSettings = Optional[Dict[str, Any]]


class CmdStatus(ABC):
    @property
    @abstractmethod
    def done(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def out_err(self) -> Tuple[str, str]:
        raise NotImplementedError


class WheelResult(NamedTuple):
    wheel: Path
    out: str
    err: str


class SdistResult(NamedTuple):
    sdist: Path
    out: str
    err: str


class MetadataForBuildWheelResult(NamedTuple):
    metadata: Path
    out: str
    err: str


class RequiresBuildWheelResult(NamedTuple):
    requires: Tuple[Requirement, ...]
    out: str
    err: str


class RequiresBuildSdistResult(NamedTuple):
    requires: Tuple[Requirement, ...]
    out: str
    err: str


class BackendFailed(RuntimeError):
    def __init__(self, result: Dict[str, Any], out: str, err: str) -> None:
        super().__init__()
        self.out = out
        self.err = err
        self.code: int = result.get("code", -2)
        self.exc_type: str = result.get("exc_type", "missing Exception type")
        self.exc_msg: str = result.get("exc_msg", "missing Exception message")

    def __str__(self) -> str:
        return (
            f"packaging backend failed{'' if self.code is None else f' (code={self.code})'}, "
            f"with {self.exc_type}: {self.exc_msg}\n{self.err}{self.out}"
        ).rstrip()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"result=dict(code={self.code}, exc_type={self.exc_type!r},exc_msg={self.exc_msg!r}),"
            f" out={self.out!r}, err={self.err!r})"
        )


class Frontend(ABC):
    LEGACY_BUILD_BACKEND: str = "setuptools.build_meta:__legacy__"
    LEGACY_REQUIRES: Tuple[Requirement, ...] = (Requirement("setuptools >= 40.8.0"), Requirement("wheel"))

    def __init__(
        self,
        root: Path,
        backend_paths: Tuple[Path, ...],
        backend_module: str,
        backend_obj: Optional[str],
        requires: Tuple[Requirement, ...],
        reuse_backend: bool = True,
    ) -> None:
        self._root = root
        self._backend_paths = backend_paths
        self._backend_module = backend_module
        self._backend_obj = backend_obj
        self._requires = requires
        self._reuse_backend = reuse_backend

    @classmethod
    def create_args_from_folder(
        cls, folder: Path
    ) -> Tuple[Path, Tuple[Path, ...], str, Optional[str], Tuple[Requirement, ...], bool]:
        py_project_toml = folder / "pyproject.toml"
        if py_project_toml.exists():
            py_project = toml.load(py_project_toml)
            build_system = py_project.get("build-system", {})
            if "backend-path" in build_system:
                backend_paths: Tuple[Path, ...] = tuple(folder / p for p in build_system["backend-path"])
            else:
                backend_paths = ()
            if "requires" in build_system:
                requires: Tuple[Requirement, ...] = tuple(Requirement(r) for r in build_system.get("requires"))
            else:
                requires = cls.LEGACY_REQUIRES
            build_backend = build_system.get("build-backend", cls.LEGACY_BUILD_BACKEND)
        else:
            backend_paths = ()
            requires = cls.LEGACY_REQUIRES
            build_backend = cls.LEGACY_BUILD_BACKEND
        paths = build_backend.split(":")
        backend_module: str = paths[0]
        backend_obj: Optional[str] = paths[1] if len(paths) > 1 else None
        return folder, backend_paths, backend_module, backend_obj, requires, True

    def build_sdist(self, sdist_directory: Path, config_settings: Optional[ConfigSettings] = None) -> SdistResult:
        sdist_directory.mkdir(parents=True, exist_ok=True)
        basename, out, err = self._send(
            cmd="build_sdist",
            sdist_directory=sdist_directory,
            config_settings=config_settings,
        )
        if not isinstance(basename, str):
            self._unexpected_response("build_sdist", basename, str, out, err)
        return SdistResult(sdist_directory / basename, out, err)

    def build_wheel(
        self,
        wheel_directory: Path,
        config_settings: Optional[ConfigSettings] = None,
        metadata_directory: Optional[Path] = None,
    ) -> WheelResult:
        wheel_directory.mkdir(parents=True, exist_ok=True)
        basename, out, err = self._send(
            cmd="build_wheel",
            wheel_directory=wheel_directory,
            config_settings=config_settings,
            metadata_directory=metadata_directory,
        )
        if not isinstance(basename, str):
            self._unexpected_response("build_wheel", basename, str, out, err)
        return WheelResult(wheel_directory / basename, out, err)

    def _unexpected_response(self, cmd: str, got: Any, expected_type: Any, out: str, err: str) -> NoReturn:
        msg = f"{cmd!r} on {self.backend!r} returned {got!r} but expected type {expected_type!r}"
        raise BackendFailed({"code": None, "exc_type": TypeError.__name__, "exc_msg": msg}, out, err)

    @property
    def backend(self) -> str:
        return f"{self._backend_module}{f':{self._backend_obj}' if self._backend_obj else ''}"

    def prepare_metadata_for_build_wheel(
        self, metadata_directory: Path, config_settings: Optional[ConfigSettings] = None
    ) -> MetadataForBuildWheelResult:
        if metadata_directory == self._root:
            raise RuntimeError(f"the project root and the metadata directory can't be the same {self._root}")
        if metadata_directory.exists():  # start with fresh
            shutil.rmtree(metadata_directory)
        metadata_directory.mkdir(parents=True, exist_ok=True)
        try:
            basename, out, err = self._send(
                cmd="prepare_metadata_for_build_wheel",
                metadata_directory=metadata_directory,
                config_settings=config_settings,
            )
        except BackendFailed:
            # if backend does not provide it acquire it from the wheel
            basename, err, out = self._metadata_from_built_wheel(config_settings, metadata_directory)
        if not isinstance(basename, str):
            self._unexpected_response("prepare_metadata_for_build_wheel", basename, str, out, err)
        result = metadata_directory / basename
        return MetadataForBuildWheelResult(result, out, err)

    def _metadata_from_built_wheel(
        self, config_settings: Optional[ConfigSettings], metadata_directory: Optional[Path]
    ) -> Tuple[str, str, str]:
        with self._wheel_directory() as wheel_directory:
            wheel_result = self.build_wheel(
                wheel_directory=wheel_directory,
                config_settings=config_settings,
                metadata_directory=metadata_directory,
            )
            wheel = wheel_result.wheel
            if not wheel.exists():
                raise RuntimeError(f"missing wheel file return by backed {wheel!r}")
            out, err = wheel_result.out, wheel_result.err
            extract_to = str(metadata_directory)
            basename = None
            with ZipFile(str(wheel), "r") as zip_file:
                for name in zip_file.namelist():
                    path = Path(name)
                    if path.parts[0].endswith(".dist-info"):
                        basename = path.parts[0]
                        zip_file.extract(name, extract_to)
            if basename is None:
                raise RuntimeError(f"no .dist-info found inside generated wheel {wheel}")
        return basename, err, out

    @contextmanager
    def _wheel_directory(self) -> Iterator[Path]:
        with TemporaryDirectory() as wheel_directory:
            yield Path(wheel_directory)

    def get_requires_for_build_wheel(
        self, config_settings: Optional[ConfigSettings] = None
    ) -> RequiresBuildWheelResult:
        try:
            result, out, err = self._send(cmd="get_requires_for_build_wheel", config_settings=config_settings)
        except BackendFailed as exc:
            result, out, err = [], exc.out, exc.err
        if not isinstance(result, list) or not all(isinstance(i, str) for i in result):
            self._unexpected_response("get_requires_for_build_wheel", result, "list of string", out, err)
        return RequiresBuildWheelResult(tuple(Requirement(r) for r in cast(List[str], result)), out, err)

    def get_requires_for_build_sdist(
        self, config_settings: Optional[ConfigSettings] = None
    ) -> RequiresBuildSdistResult:
        try:
            result, out, err = self._send(cmd="get_requires_for_build_sdist", config_settings=config_settings)
        except BackendFailed as exc:
            result, out, err = [], exc.out, exc.err
        if not isinstance(result, list) or not all(isinstance(i, str) for i in result):
            self._unexpected_response("get_requires_for_build_sdist", result, "list of string", out, err)
        return RequiresBuildSdistResult(tuple(Requirement(r) for r in cast(List[str], result)), out, err)

    def _send(self, cmd: str, **kwargs: Any) -> Tuple[Any, str, str]:
        with NamedTemporaryFile(prefix=f"pep517_{cmd}-") as result_file_marker:
            result_file = Path(result_file_marker.name).with_suffix(".json")
            msg = json.dumps(
                {
                    "cmd": cmd,
                    "kwargs": {k: (str(v) if isinstance(v, Path) else v) for k, v in kwargs.items()},
                    "result": str(result_file),
                }
            )
            with self._send_msg(cmd, result_file, msg) as status:
                while not status.done:
                    sleep(0.001)  # wait a bit for things to happen
            if result_file.exists():
                try:
                    with result_file.open("rt") as result_handler:
                        result = json.load(result_handler)
                finally:
                    result_file.unlink()
            else:
                result = {
                    "code": 1,
                    "exc_type": "RuntimeError",
                    "exc_msg": f"Backend response file {result_file} is missing",
                }
        out, err = status.out_err()
        if "return" in result:
            return result["return"], out, err
        raise BackendFailed(result, out, err)

    @property
    def backend_args(self) -> List[str]:
        result: List[str] = [str(_HERE / "backend.py"), str(self._reuse_backend), self._backend_module]
        if self._backend_obj:
            result.append(self._backend_obj)
        return result

    @abstractmethod
    @contextmanager
    def _send_msg(self, cmd: str, result_file: Path, msg: str) -> Iterator[CmdStatus]:
        raise NotImplementedError



from pathlib import Path
from typing import List, Set, Tuple, cast

from packaging.requirements import Requirement

from tox.plugin.impl import impl
from tox.tox_env.register import ToxEnvRegister

from ..api import Pep517VirtualEnvPackage


class LegacyDevVirtualEnvPackage(Pep517VirtualEnvPackage):
    """Use PEP-517 to get package build and runtime dependencies - install the root folder"""

    def _build_artifact(self) -> Path:
        return cast(Path, self.core["tox_root"])  # the folder itself is the package

    def get_package_dependencies(self, extras: Set[str]) -> List[Requirement]:
        # install build-requires dependencies so that the legacy installer has them satisfied when installing package
        result = super().get_package_dependencies(extras)
        result.extend(self.build_requires())
        return result

    @staticmethod
    def id() -> str:
        return "virtualenv-legacy-dev"

    def build_requires(self) -> Tuple[Requirement, ...]:
        result: Tuple[Requirement, ...] = ()
        return result


@impl
def tox_register_tox_env(register: ToxEnvRegister) -> None:
    register.add_package_env(LegacyDevVirtualEnvPackage)



import sys

import pytest

from tests.config.loader.ini.replace.conftest import ReplaceOne


def test_replace_pos_args_none_sys_argv(replace_one: ReplaceOne) -> None:
    result = replace_one("{posargs}", None)
    assert result == ""


def test_replace_pos_args_empty_sys_argv(replace_one: ReplaceOne) -> None:
    result = replace_one("{posargs}", None)
    assert result == ""


def test_replace_pos_args_extra_sys_argv(replace_one: ReplaceOne) -> None:
    result = replace_one("{posargs}", [sys.executable, "magic"])
    assert result == f"{sys.executable} magic"


def test_replace_pos_args(replace_one: ReplaceOne) -> None:
    result = replace_one("{posargs}", ["ok", "what", " yes "])
    quote = '"' if sys.platform == "win32" else "'"
    assert result == f"ok what {quote} yes {quote}"


@pytest.mark.parametrize(
    ["value", "result"],
    [
        ("magic", "magic"),
        ("magic:colon", "magic:colon"),
        ("magic\n b:c", "magic\nb:c"),  # unescaped newline keeps the newline
        ("magi\\\n c:d", "magic:d"),  # escaped newline merges the lines
        ("\\{a\\}", "{a}"),  # escaped curly braces
    ],
)
def test_replace_pos_args_default(replace_one: ReplaceOne, value: str, result: str) -> None:
    """If we have a factor that is not specified within the core env-list then that's also an environment"""
    outcome = replace_one(f"{{posargs:{value}}}", None)
    assert result == outcome


@pytest.mark.parametrize(
    "value",
    [
        "\\{posargs}",
        "{posargs\\}",
        "\\{posargs\\}",
        "{\\{posargs}",
        "{\\{posargs}{}",
    ],
)
def test_replace_pos_args_escaped(replace_one: ReplaceOne, value: str) -> None:
    result = replace_one(value, None)
    outcome = value.replace("\\{", "{").replace("\\}", "}")
    assert result == outcome



"""A executor that reuses a single subprocess for all backend calls (saving on python startup/import overhead)"""
from pathlib import Path
from subprocess import TimeoutExpired
from threading import Lock
from types import TracebackType
from typing import Dict, Optional, Sequence, Tuple, Type

from tox.execute import ExecuteRequest
from tox.execute.api import Execute, ExecuteInstance, ExecuteStatus
from tox.execute.local_sub_process import LocalSubProcessExecuteInstance
from tox.execute.request import StdinSource
from tox.execute.stream import SyncWrite
from tox.util.pep517.frontend import BackendFailed


class LocalSubProcessPep517Executor(Execute):
    """Executor holding the backend process"""

    def __init__(self, colored: bool, cmd: Sequence[str], env: Dict[str, str], cwd: Path):
        super().__init__(colored)
        self.cmd = cmd
        self.env = env
        self.cwd = cwd
        self._local_execute: Optional[Tuple[LocalSubProcessExecuteInstance, ExecuteStatus]] = None
        self._exc: Optional[Exception] = None
        self.is_alive: bool = False

    def build_instance(self, request: ExecuteRequest, out: SyncWrite, err: SyncWrite) -> ExecuteInstance:
        result = LocalSubProcessPep517ExecuteInstance(request, out, err, self.local_execute)
        return result

    @property
    def local_execute(self) -> Tuple[LocalSubProcessExecuteInstance, ExecuteStatus]:
        if self._exc is not None:
            raise self._exc
        if self._local_execute is None:
            request = ExecuteRequest(cmd=self.cmd, cwd=self.cwd, env=self.env, stdin=StdinSource.API, run_id="pep517")

            instance = LocalSubProcessExecuteInstance(
                request,
                out=SyncWrite(name="pep517-out", target=None, color=None),  # not enabled no need to enter/exit
                err=SyncWrite(name="pep517-err", target=None, color=None),  # not enabled no need to enter/exit
                on_exit_drain=False,
            )
            status = instance.__enter__()
            self._local_execute = instance, status
            while True:
                if b"started backend " in status.out:
                    self.is_alive = True
                    break
                if b"failed to start backend" in status.err:
                    from tox.tox_env.python.virtual_env.package.api import ToxBackendFailed

                    failure = BackendFailed(
                        result={
                            "code": -5,
                            "exc_type": "FailedToStart",
                            "exc_msg": "could not start backend",
                        },
                        out=status.out.decode(),
                        err=status.err.decode(),
                    )
                    self._exc = ToxBackendFailed(failure)
                    raise self._exc
        return self._local_execute

    @staticmethod
    def _handler(into: bytearray, content: bytes) -> None:
        """ignore content generated"""
        into.extend(content)  # pragma: no cover

    def close(self) -> None:
        if self._local_execute is not None:  # pragma: no branch
            execute, status = self._local_execute
            execute.__exit__(None, None, None)
            if execute.process is not None:  # pragma: no branch
                if execute.process.returncode is None:  # pragma: no cover
                    try:  # pragma: no cover
                        execute.process.wait(timeout=0.1)  # pragma: no cover
                    except TimeoutExpired:  # pragma: no cover
                        execute.process.terminate()  # pragma: no cover  # if does not stop on its own kill it


class LocalSubProcessPep517ExecuteInstance(ExecuteInstance):
    """A backend invocation"""

    def __init__(
        self,
        request: ExecuteRequest,
        out: SyncWrite,
        err: SyncWrite,
        instance_status: Tuple[LocalSubProcessExecuteInstance, ExecuteStatus],
    ):
        super().__init__(request, out, err)
        self._instance, self._status = instance_status
        self._lock = Lock()

    @property
    def cmd(self) -> Sequence[str]:
        return self._instance.cmd

    def __enter__(self) -> ExecuteStatus:
        self._lock.acquire()
        self._swap_out_err()
        return self._status

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        self._swap_out_err()
        self._lock.release()

    def _swap_out_err(self) -> None:
        out, err = self._out, self._err
        # update status to see the newly collected content
        self._out, self._err = self._instance.set_out_err(out, err)
        # update the thread out/err
        self._status.set_out_err(out, err)



from typing import Optional

import pytest

from tox.pytest import ToxProjectCreator


def test_devenv_fail_multiple_target(tox_project: ToxProjectCreator) -> None:
    outcome = tox_project({"tox.ini": ""}).run("d", "-e", "py39,py38")
    outcome.assert_failed()
    msg = "ROOT: HandledError| exactly one target environment allowed in devenv mode but found py39, py38\n"
    outcome.assert_out_err(msg, "")


@pytest.mark.integration
@pytest.mark.timeout(60)
def test_devenv_ok(tox_project: ToxProjectCreator, enable_pip_pypi_access: Optional[str]) -> None:
    content = {"setup.py": "from setuptools import setup\nsetup(name='demo', version='1.0')"}
    outcome = tox_project(content).run("d", "-e", "py")
    outcome.assert_success()



import json
import logging
import os
import subprocess
import sys
from io import TextIOWrapper
from pathlib import Path
from typing import Dict, List, Tuple

import psutil
import pytest
from colorama import Fore
from psutil import AccessDenied
from pytest_mock import MockerFixture

from tox.execute.api import Outcome
from tox.execute.local_sub_process import SIG_INTERRUPT, LocalSubProcessExecutor
from tox.execute.request import ExecuteRequest, StdinSource
from tox.pytest import CaptureFixture, LogCaptureFixture, MonkeyPatch
from tox.report import NamedBytesIO


class FakeOutErr:
    def __init__(self) -> None:
        self.out_err = TextIOWrapper(NamedBytesIO("out")), TextIOWrapper(NamedBytesIO("err"))

    def read_out_err(self) -> Tuple[str, str]:
        out_got = self.out_err[0].buffer.getvalue().decode(self.out_err[0].encoding)  # type: ignore[attr-defined]
        err_got = self.out_err[1].buffer.getvalue().decode(self.out_err[1].encoding)  # type: ignore[attr-defined]
        return out_got, err_got


@pytest.mark.parametrize("color", [True, False], ids=["color", "no_color"])
@pytest.mark.parametrize(["out", "err"], [("out", "err"), ("", "")], ids=["simple", "nothing"])
@pytest.mark.parametrize("show", [True, False], ids=["show", "no_show"])
def test_local_execute_basic_pass(
    caplog: LogCaptureFixture,
    os_env: Dict[str, str],
    out: str,
    err: str,
    show: bool,
    color: bool,
) -> None:
    caplog.set_level(logging.NOTSET)
    executor = LocalSubProcessExecutor(colored=color)
    code = f"import sys; print({repr(out)}, end=''); print({repr(err)}, end='', file=sys.stderr)"
    request = ExecuteRequest(cmd=[sys.executable, "-c", code], cwd=Path(), env=os_env, stdin=StdinSource.OFF, run_id="")
    out_err = FakeOutErr()
    with executor.call(request, show=show, out_err=out_err.out_err) as status:
        while status.exit_code is None:
            status.wait()
    assert status.out == out.encode()
    assert status.err == err.encode()
    outcome = status.outcome
    assert outcome is not None
    assert bool(outcome) is True, outcome
    assert outcome.exit_code == Outcome.OK
    assert outcome.err == err
    assert outcome.out == out
    assert outcome.request == request

    out_got, err_got = out_err.read_out_err()
    if show:
        assert out_got == out
        expected = (f"{Fore.RED}{err}{Fore.RESET}" if color else err) if err else ""
        assert err_got == expected
    else:
        assert not out_got
        assert not err_got
    assert not caplog.records


def test_local_execute_basic_pass_show_on_standard_newline_flush(caplog: LogCaptureFixture) -> None:
    caplog.set_level(logging.NOTSET)
    executor = LocalSubProcessExecutor(colored=False)
    request = ExecuteRequest(
        cmd=[sys.executable, "-c", "import sys; print('out'); print('yay')"],
        cwd=Path(),
        env=os.environ.copy(),
        stdin=StdinSource.OFF,
        run_id="",
    )
    out_err = FakeOutErr()
    with executor.call(request, show=True, out_err=out_err.out_err) as status:
        while status.exit_code is None:
            status.wait()
    outcome = status.outcome
    assert outcome is not None
    assert repr(outcome)
    assert bool(outcome) is True, outcome
    assert outcome.exit_code == Outcome.OK
    assert not outcome.err
    assert outcome.out == f"out{os.linesep}yay{os.linesep}"
    out, err = out_err.read_out_err()
    assert out == f"out{os.linesep}yay{os.linesep}"
    assert not err
    assert not caplog.records


def test_local_execute_write_a_lot(caplog: LogCaptureFixture, os_env: Dict[str, str]) -> None:
    count = 10_000
    executor = LocalSubProcessExecutor(colored=False)
    request = ExecuteRequest(
        cmd=[
            sys.executable,
            "-c",
            (
                "import sys; import time; from datetime import datetime; import os;"
                "print('e' * {0}, file=sys.stderr);"
                "print('o' * {0}, file=sys.stdout);"
                "time.sleep(0.5);"
                "print('a' * {0}, file=sys.stderr);"
                "print('b' * {0}, file=sys.stdout);"
            ).format(count),
        ],
        cwd=Path(),
        env=os_env,
        stdin=StdinSource.OFF,
        run_id="",
    )
    out_err = FakeOutErr()
    with executor.call(request, show=False, out_err=out_err.out_err) as status:
        while status.exit_code is None:
            status.wait()
    outcome = status.outcome
    assert outcome is not None
    assert bool(outcome), outcome
    expected_out = f"{'o' * count}{os.linesep}{'b' * count}{os.linesep}"
    assert outcome.out == expected_out, expected_out[len(outcome.out) :]
    expected_err = f"{'e' * count}{os.linesep}{'a' * count}{os.linesep}"
    assert outcome.err == expected_err, expected_err[len(outcome.err) :]


def test_local_execute_basic_fail(capsys: CaptureFixture, caplog: LogCaptureFixture, monkeypatch: MonkeyPatch) -> None:
    monkeypatch.chdir(Path(__file__).parents[3])
    caplog.set_level(logging.NOTSET)
    executor = LocalSubProcessExecutor(colored=False)
    cwd = Path().absolute()
    cmd = [
        sys.executable,
        "-c",
        "import sys; print('out', end=''); print('err', file=sys.stderr, end=''); sys.exit(3)",
    ]
    request = ExecuteRequest(cmd=cmd, cwd=cwd, env=os.environ.copy(), stdin=StdinSource.OFF, run_id="")

    # run test
    out_err = FakeOutErr()
    with executor.call(request, show=False, out_err=out_err.out_err) as status:
        while status.exit_code is None:
            status.wait()
    outcome = status.outcome
    assert outcome is not None

    assert repr(outcome)

    # assert no output, no logs
    out, err = out_err.read_out_err()
    assert not out
    assert not err
    assert not caplog.records

    # assert return object
    assert bool(outcome) is False, outcome
    assert outcome.exit_code == 3
    assert outcome.err == "err"
    assert outcome.out == "out"
    assert outcome.request == request

    # asset fail
    with pytest.raises(SystemExit) as context:
        outcome.assert_success()
    # asset fail
    assert context.value.code == 3

    out, err = capsys.readouterr()
    assert out == "out\n"
    expected = f"{Fore.RED}err{Fore.RESET}\n"
    assert err == expected

    assert len(caplog.records) == 1
    record = caplog.records[0]
    assert record.levelno == logging.CRITICAL
    assert record.msg == "exit %s (%.2f seconds) %s> %s"
    _code, _duration, _cwd, _cmd = record.args
    assert _code == 3
    assert _cwd == cwd
    assert _cmd == request.shell_cmd
    assert isinstance(_duration, float)
    assert _duration > 0


def test_command_does_not_exist(capsys: CaptureFixture, caplog: LogCaptureFixture, os_env: Dict[str, str]) -> None:
    caplog.set_level(logging.NOTSET)
    executor = LocalSubProcessExecutor(colored=False)
    request = ExecuteRequest(
        cmd=["sys-must-be-missing"], cwd=Path().absolute(), env=os_env, stdin=StdinSource.OFF, run_id=""
    )
    out_err = FakeOutErr()
    with executor.call(request, show=False, out_err=out_err.out_err) as status:
        while status.exit_code is None:  # pragma: no branch
            status.wait()  # pragma: no cover
    outcome = status.outcome
    assert outcome is not None

    assert bool(outcome) is False, outcome
    assert outcome.exit_code != Outcome.OK
    assert outcome.out == ""
    assert outcome.err == ""
    assert not caplog.records


@pytest.mark.skipif(sys.platform == "win32", reason="You need a conhost shell for keyboard interrupt")
def test_command_keyboard_interrupt(tmp_path: Path, monkeypatch: MonkeyPatch, capfd: CaptureFixture) -> None:
    monkeypatch.chdir(tmp_path)
    process_up_signal = tmp_path / "signal"
    cmd = [sys.executable, str(Path(__file__).parent / "local_subprocess_sigint.py"), str(process_up_signal)]
    process = subprocess.Popen(cmd)
    while not process_up_signal.exists():
        assert process.poll() is None
    root = process.pid
    try:
        child = next(iter(psutil.Process(pid=root).children())).pid
    except AccessDenied as exc:  # pragma: no cover # on termux for example
        pytest.skip(str(exc))  # pragma: no cover

    print(f"test running in {os.getpid()} and sending CTRL+C to {process.pid}", file=sys.stderr)
    process.send_signal(SIG_INTERRUPT)
    try:
        process.communicate(timeout=3)
    except subprocess.TimeoutExpired:  # pragma: no cover
        process.kill()
        raise

    out, err = capfd.readouterr()
    assert f"W	requested interrupt of {child} from {root}" in err, err
    assert f"W	send signal SIGINT(2) to {child} from {root} with timeout 0.30" in err, err
    assert f"W	send signal SIGTERM(15) to {child} from {root} with timeout 0.20" in err, err
    assert f"W	send signal SIGKILL(9) to {child} from {root}" in err, err

    outs = out.split("\n")

    exit_code = int(outs[0])
    assert exit_code == -9
    assert float(outs[3]) > 0  # duration
    assert "how about no signal 2" in outs[1], outs[1]  # 2 - Interrupt
    assert "how about no signal 15" in outs[1], outs[1]  # 15 - Terminated


@pytest.mark.parametrize("tty_mode", ["on", "off"])
def test_local_subprocess_tty(monkeypatch: MonkeyPatch, mocker: MockerFixture, tty_mode: str) -> None:
    monkeypatch.setenv("COLUMNS", "100")
    monkeypatch.setenv("LINES", "100")
    tty = tty_mode == "on"
    mocker.patch("sys.stdout.isatty", return_value=tty)
    mocker.patch("sys.stderr.isatty", return_value=tty)

    executor = LocalSubProcessExecutor(colored=False)
    cmd: List[str] = [sys.executable, str(Path(__file__).parent / "tty_check.py")]
    request = ExecuteRequest(cmd=cmd, stdin=StdinSource.API, cwd=Path.cwd(), env=dict(os.environ), run_id="")
    out_err = FakeOutErr()
    with executor.call(request, show=False, out_err=out_err.out_err) as status:
        while status.exit_code is None:
            status.wait()
    outcome = status.outcome
    assert outcome is not None

    assert outcome
    info = json.loads(outcome.out)
    assert info == {
        "stdout": False,
        "stderr": False,
        "stdin": False,
        "terminal": [100, 100],
    }



from tox.pytest import ToxProjectCreator


def test_env_already_packaging(tox_project: ToxProjectCreator) -> None:
    proj = tox_project(
        {
            "tox.ini": "[testenv]\npackage=wheel",
            "pyproject.toml": '[build-system]\nrequires=[]\nbuild-backend="build"',
        }
    )

    result = proj.run("r", "-e", "py,.package-py")
    result.assert_failed(code=-2)
    assert "cannot run packaging environment .package-py" in result.out, result.out


def test_env_run_cannot_be_packaging_too(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({"tox.ini": "[testenv]\npackage=wheel\npackage_env=py"})
    result = proj.run("r", "-e", "py")
    result.assert_failed(code=-2)
    assert " py is already defined as a run environment, cannot be packaging too" in result.out, result.out



"""
Declare the abstract base class for tox environments that handle the Python language via the virtualenv project.
"""
from abc import ABC
from pathlib import Path
from typing import Dict, List, Optional, Sequence, cast

from virtualenv import session_via_cli
from virtualenv.create.creator import Creator
from virtualenv.run.session import Session

from tox.config.cli.parser import DEFAULT_VERBOSITY, Parsed
from tox.config.sets import ConfigSet
from tox.execute.api import Execute, Outcome, StdinSource
from tox.execute.local_sub_process import LocalSubProcessExecutor
from tox.journal import EnvJournal
from tox.report import ToxHandler

from ..api import Python, PythonDeps, PythonInfo


class VirtualEnv(Python, ABC):
    """A python executor that uses the virtualenv project with pip"""

    def __init__(
        self, conf: ConfigSet, core: ConfigSet, options: Parsed, journal: EnvJournal, log_handler: ToxHandler
    ) -> None:
        self._virtualenv_session: Optional[Session] = None  # type: ignore[no-any-unimported]
        super().__init__(conf, core, options, journal, log_handler)

    def default_pass_env(self) -> List[str]:
        env = super().default_pass_env()
        env.append("PIP_*")  # we use pip as installer
        env.append("VIRTUALENV_*")  # we use virtualenv as isolation creator
        return env

    def default_set_env(self) -> Dict[str, str]:
        env = super().default_set_env()
        env["PIP_DISABLE_PIP_VERSION_CHECK"] = "1"
        env["VIRTUALENV_NO_PERIODIC_UPDATE"] = "1"
        return env

    def build_executor(self) -> Execute:
        return LocalSubProcessExecutor(self.options.is_colored)

    @property
    def session(self) -> Session:  # type: ignore[no-any-unimported]
        if self._virtualenv_session is None:
            args = [
                "--clear",
                "--no-periodic-update",
                str(cast(Path, self.conf["env_dir"])),
            ]
            base_python: List[str] = self.conf["base_python"]
            for base in base_python:
                args.extend(["-p", base])
            self._virtualenv_session = session_via_cli(args, setup_logging=False)
        return self._virtualenv_session

    @property
    def creator(self) -> Creator:  # type: ignore[no-any-unimported]
        return self.session.creator

    def create_python_env(self) -> None:
        self.session.run()

    def _get_python(self, base_python: List[str]) -> Optional[PythonInfo]:
        try:
            interpreter = self.creator.interpreter
        except RuntimeError:  # if can't find
            return None
        return PythonInfo(
            executable=Path(interpreter.system_executable),
            implementation=interpreter.implementation,
            version_info=interpreter.version_info,
            version=interpreter.version,
            is_64=(interpreter.architecture == 64),
            platform=interpreter.platform,
            extra_version_info=None,
        )

    def paths(self) -> List[Path]:
        """Paths to add to the executable"""
        # we use the original executable as shims may be somewhere else
        return list({self.creator.bin_dir, self.creator.script_dir})

    def env_site_package_dir(self) -> Path:
        return cast(Path, self.creator.purelib)

    def env_python(self) -> Path:
        return cast(Path, self.creator.exe)

    def install_python_packages(
        self,
        packages: PythonDeps,
        of_type: str,
        no_deps: bool = False,
        develop: bool = False,
        force_reinstall: bool = False,
    ) -> None:
        if not packages:
            return
        install_command = self.base_install_cmd
        if no_deps:
            install_command.append("--no-deps")
        if force_reinstall:
            install_command.append("--force-reinstall")
        if develop is True:
            install_command.extend(("--no-build-isolation", "-e"))
        install_command.extend(str(i) for i in packages)
        result = self.perform_install(install_command, f"install_{of_type}")
        result.assert_success()

    @property
    def base_install_cmd(self) -> List[str]:
        return [str(self.creator.exe), "-I", "-m", "pip", "install"]

    def perform_install(self, install_command: Sequence[str], run_id: str) -> Outcome:
        return self.execute(
            cmd=install_command,
            stdin=StdinSource.OFF,
            cwd=self.core["tox_root"],
            run_id=run_id,
            show=self.options.verbosity > DEFAULT_VERBOSITY,
        )

    def get_installed_packages(self) -> List[str]:
        list_command = [self.creator.exe, "-I", "-m", "pip", "freeze", "--all"]
        result = self.execute(
            cmd=list_command,
            stdin=StdinSource.OFF,
            run_id="freeze",
            show=self.options.verbosity > DEFAULT_VERBOSITY,
        )
        result.assert_success()
        return result.out.splitlines()



"""Load """
from configparser import ConfigParser
from itertools import chain
from pathlib import Path
from typing import Dict, Iterator, List, Optional, Set

from tox.config.loader.ini.factor import find_envs

from ..loader.api import OverrideMap
from ..loader.ini import IniLoader
from ..loader.ini.replace import BASE_TEST_ENV, CORE_PREFIX
from ..sets import ConfigSet
from . import Source

TEST_ENV_PREFIX = f"{BASE_TEST_ENV}:"


class ToxIni(Source):
    """Configuration sourced from a ini file (such as tox.ini)"""

    def __init__(self, path: Path) -> None:
        self._path = path
        self._parser = ConfigParser()
        if self._path.exists():  # if does not exist keep it empty
            with self._path.open() as file_handler:
                self._parser.read_file(file_handler)

        super().__init__()
        self._envs: Dict[Optional[str], List[IniLoader]] = {}

    def get_core(self, override_map: OverrideMap) -> Iterator[IniLoader]:
        if None in self._envs:
            yield from self._envs[None]
            return
        core = []
        if self._parser.has_section(CORE_PREFIX):
            core.append(
                IniLoader(
                    section=CORE_PREFIX,
                    parser=self._parser,
                    overrides=override_map.get(CORE_PREFIX, []),
                )
            )
        self._envs[None] = core
        yield from core

    def get_env_loaders(
        self, env_name: str, override_map: OverrideMap, package: bool, conf: ConfigSet
    ) -> Iterator[IniLoader]:
        section = f"{TEST_ENV_PREFIX}{env_name}"
        try:
            yield from self._envs[section]
        except KeyError:
            loaders: List[IniLoader] = []
            self._envs[section] = loaders

            if self._parser.has_section(section):
                loader = IniLoader(
                    section=section,
                    parser=self._parser,
                    overrides=override_map.get(section, []),
                )
                yield loader
                loaders.append(loader)

            if package is False:
                conf.add_config(  # base may be override within the testenv:py section
                    keys="base",
                    of_type=List[str],
                    desc="inherit missing keys from these sections",
                    default=[BASE_TEST_ENV],
                )
                for base in conf["base"]:
                    for section in (base, f"{TEST_ENV_PREFIX}{base}"):
                        if self._parser.has_section(section):
                            loader = IniLoader(
                                section=section,
                                parser=self._parser,
                                overrides=override_map.get(section, []),
                            )
                            yield loader
                            loaders.append(loader)
                            break

    def envs(self, core_config: ConfigSet) -> Iterator[str]:
        seen = set()
        for name in self._discover_tox_envs(core_config):
            if name not in seen:
                seen.add(name)
                yield name

    def _discover_tox_envs(self, core_config: ConfigSet) -> Iterator[str]:
        explicit = list(core_config["env_list"])
        yield from explicit
        known_factors = None
        for section in self._parser.sections():
            if section.startswith(BASE_TEST_ENV):
                is_base_section = section == BASE_TEST_ENV
                name = BASE_TEST_ENV if is_base_section else section[len(TEST_ENV_PREFIX) :]
                if not is_base_section:
                    yield name
                if known_factors is None:
                    known_factors = set(chain.from_iterable(e.split("-") for e in explicit))
                yield from self._discover_from_section(section, known_factors)

    def _discover_from_section(self, section: str, known_factors: Set[str]) -> Iterator[str]:
        for value in self._parser[section].values():
            for env in find_envs(value):
                if env not in known_factors:
                    yield env

    def __repr__(self) -> str:
        return f"{type(self).__name__}(path={self._path})"


__all__ = ("ToxIni",)



from typing import TYPE_CHECKING, Dict, Iterable, Type

from .package import PackageToxEnv
from .runner import RunToxEnv

if TYPE_CHECKING:
    from tox.plugin.manager import Plugin


class ToxEnvRegister:
    def __init__(self) -> None:
        self._run_envs: Dict[str, Type[RunToxEnv]] = {}
        self._package_envs: Dict[str, Type[PackageToxEnv]] = {}
        self._default_run_env: str = ""

    def populate(self, manager: "Plugin") -> None:
        manager.tox_register_tox_env(register=self)
        self._default_run_env = next(iter(self._run_envs.keys()))

    def add_run_env(self, of_type: Type[RunToxEnv]) -> None:
        self._run_envs[of_type.id()] = of_type

    def add_package_env(self, of_type: Type[PackageToxEnv]) -> None:
        self._package_envs[of_type.id()] = of_type

    @property
    def run_envs(self) -> Iterable[str]:
        return self._run_envs.keys()

    @property
    def default_run_env(self) -> str:
        return self._default_run_env

    @default_run_env.setter
    def default_run_env(self, value: str) -> None:
        if value not in self._run_envs:
            raise ValueError("run env must be registered before setting it as default")
        self._default_run_env = value

    def runner(self, name: str) -> Type[RunToxEnv]:
        return self._run_envs[name]

    def package(self, name: str) -> Type[PackageToxEnv]:
        return self._package_envs[name]


REGISTER = ToxEnvRegister()



"""
A pytest plugin useful to test tox itself (and its plugins).
"""

import os
import random
import re
import shutil
import socket
import string
import sys
import textwrap
import warnings
from contextlib import closing, contextmanager
from pathlib import Path
from subprocess import PIPE, Popen, check_call
from threading import Thread
from types import TracebackType
from typing import IO, TYPE_CHECKING, Any, Callable, Dict, Iterator, List, Optional, Sequence, Tuple, Type, cast
from unittest.mock import MagicMock

import pytest
from _pytest.capture import CaptureFixture as _CaptureFixture
from _pytest.config import Config as PyTestConfig
from _pytest.config.argparsing import Parser
from _pytest.logging import LogCaptureFixture
from _pytest.monkeypatch import MonkeyPatch
from _pytest.python import Function
from _pytest.tmpdir import TempPathFactory
from pytest_mock import MockerFixture
from virtualenv.discovery.py_info import PythonInfo
from virtualenv.info import IS_WIN, fs_supports_symlink

import tox.run
from tox.execute.api import Execute, ExecuteInstance, ExecuteStatus, Outcome
from tox.execute.request import ExecuteRequest, shell_cmd
from tox.execute.stream import SyncWrite
from tox.report import LOGGER, OutErr
from tox.run import run as tox_run
from tox.run import setup_state as previous_setup_state
from tox.session.cmd.run.parallel import ENV_VAR_KEY
from tox.session.state import State
from tox.tox_env import api as tox_env_api
from tox.tox_env.api import ToxEnv

if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
    from typing import Protocol
else:  # pragma: no cover (<py38)
    from typing_extensions import Protocol  # noqa

if TYPE_CHECKING:
    CaptureFixture = _CaptureFixture[str]
else:
    CaptureFixture = _CaptureFixture

os.environ["PIP_DISABLE_PIP_VERSION_CHECK"] = "1"
os.environ["PIP_NO_PYTHON_VERSION_WARNING"] = "1"

if fs_supports_symlink():  # pragma: no cover # used to speed up test suite run time where possible
    os.environ["VIRTUALENV_SYMLINK_APP_DATA"] = "1"
    os.environ["VIRTUALENV_SYMLINKS"] = "1"


@pytest.fixture(autouse=True)
def ensure_logging_framework_not_altered() -> Iterator[None]:
    before_handlers = list(LOGGER.handlers)
    yield
    LOGGER.handlers = before_handlers


@contextmanager
def check_os_environ() -> Iterator[None]:
    old = os.environ.copy()
    to_clean = {k: os.environ.pop(k, None) for k in {ENV_VAR_KEY, "TOX_WORK_DIR", "PYTHONPATH", "COV_CORE_CONTEXT"}}

    yield

    for key, value in to_clean.items():
        if value is not None:
            os.environ[key] = value

    new = os.environ
    extra = {k: new[k] for k in set(new) - set(old)}
    extra.pop("PLAT", None)
    miss = {k: old[k] for k in set(old) - set(new)}
    diff = {
        f"{k} = {old[k]} vs {new[k]}" for k in set(old) & set(new) if old[k] != new[k] and not k.startswith("PYTEST_")
    }
    if extra or miss or diff:
        msg = "test changed environ"
        if extra:
            msg += f" extra {extra}"
        if miss:
            msg += f" miss {miss}"
        if diff:
            msg += f" diff {diff}"
        pytest.fail(msg)


@pytest.fixture(autouse=True)
def check_os_environ_stable(monkeypatch: MonkeyPatch) -> Iterator[None]:
    with check_os_environ():
        yield
        monkeypatch.undo()


@pytest.fixture(autouse=True)
def no_color(monkeypatch: MonkeyPatch, check_os_environ_stable: None) -> None:
    monkeypatch.setenv("NO_COLOR", "yes")


class ToxProject:
    def __init__(
        self,
        files: Dict[str, Any],
        base: Optional[Path],
        path: Path,
        capfd: CaptureFixture,
        monkeypatch: MonkeyPatch,
        mocker: MockerFixture,
    ) -> None:
        self.path: Path = path
        self.monkeypatch: MonkeyPatch = monkeypatch
        self.mocker = mocker
        self._capfd = capfd
        self._setup_files(self.path, base, files)

    @staticmethod
    def _setup_files(dest: Path, base: Optional[Path], content: Dict[str, Any]) -> None:
        if base is not None:
            shutil.copytree(str(base), str(dest))
        dest.mkdir(exist_ok=True)
        for key, value in content.items():
            if not isinstance(key, str):
                raise TypeError(f"{key!r} at {dest}")  # pragma: no cover
            at_path = dest / key
            if isinstance(value, dict):
                at_path.mkdir(exist_ok=True)
                ToxProject._setup_files(at_path, None, value)
            elif isinstance(value, str):
                at_path.write_text(textwrap.dedent(value))
            else:
                msg = "could not handle {} with content {!r}".format(at_path / key, value)  # pragma: no cover
                raise TypeError(msg)  # pragma: no cover

    def patch_execute(self, handle: Callable[[ExecuteRequest], Optional[int]]) -> MagicMock:
        class MockExecute(Execute):
            def __init__(self, colored: bool, exit_code: int) -> None:
                self.exit_code = exit_code
                super().__init__(colored)

            def build_instance(self, request: ExecuteRequest, out: SyncWrite, err: SyncWrite) -> ExecuteInstance:
                return MockExecuteInstance(request, out, err, self.exit_code)

        class MockExecuteStatus(ExecuteStatus):
            def __init__(self, out: SyncWrite, err: SyncWrite, exit_code: int) -> None:
                super().__init__(out, err)
                self._exit_code = exit_code

            @property
            def exit_code(self) -> Optional[int]:
                return self._exit_code

            def wait(self, timeout: Optional[float] = None) -> None:
                """"""

            def write_stdin(self, content: str) -> None:
                """"""

            def interrupt(self) -> None:
                """"""

        class MockExecuteInstance(ExecuteInstance):
            def __init__(self, request: ExecuteRequest, out: SyncWrite, err: SyncWrite, exit_code: int) -> None:
                super().__init__(request, out, err)
                self.exit_code = exit_code

            def __enter__(self) -> ExecuteStatus:
                return MockExecuteStatus(self._out, self._err, self.exit_code)

            def __exit__(
                self,
                exc_type: Optional[Type[BaseException]],
                exc_val: Optional[BaseException],
                exc_tb: Optional[TracebackType],
            ) -> None:
                pass

            @property
            def cmd(self) -> Sequence[str]:
                return self.request.cmd

        @contextmanager
        def _execute_call(
            executor: Execute, out_err: OutErr, request: ExecuteRequest, show: bool
        ) -> Iterator[ExecuteStatus]:
            exit_code = handle(request)
            if exit_code is not None:
                executor = MockExecute(colored=executor._colored, exit_code=exit_code)  # noqa
            with original_execute_call(executor, out_err, request, show) as status:
                yield status

        original_execute_call = ToxEnv._execute_call  # noqa
        return self.mocker.patch.object(ToxEnv, "_execute_call", side_effect=_execute_call)

    @property
    def structure(self) -> Dict[str, Any]:
        result: Dict[str, Any] = {}
        for dir_name, _, files in os.walk(str(self.path)):
            dir_path = Path(dir_name)
            into = result
            relative = dir_path.relative_to(str(self.path))
            for elem in relative.parts:
                into = into.setdefault(elem, {})
            for file_name in files:
                into[file_name] = (dir_path / file_name).read_text()
        return result

    @contextmanager
    def chdir(self) -> Iterator[None]:
        cur_dir = os.getcwd()
        os.chdir(str(self.path))
        try:
            yield
        finally:
            os.chdir(cur_dir)

    def run(self, *args: str) -> "ToxRunOutcome":
        with self.chdir():
            state = None
            self._capfd.readouterr()  # start with a clean state - drain
            code = None
            state = None

            def our_setup_state(value: Sequence[str]) -> State:
                nonlocal state
                state = previous_setup_state(value)
                return state

            with self.monkeypatch.context() as m:
                m.setattr(tox_env_api, "_CWD", self.path)
                m.setattr(tox.run, "setup_state", our_setup_state)
                m.setattr(sys, "argv", [sys.executable, "-m", "tox"] + list(args))
                m.setenv("VIRTUALENV_SYMLINK_APP_DATA", "1")
                m.setenv("VIRTUALENV_SYMLINKS", "1")
                m.setenv("VIRTUALENV_PIP", "embed")
                m.setenv("VIRTUALENV_WHEEL", "embed")
                m.setenv("VIRTUALENV_SETUPTOOLS", "embed")
                try:
                    tox_run(args)
                except SystemExit as exception:
                    code = exception.code
                if code is None:
                    raise RuntimeError("exit code not set")
            out, err = self._capfd.readouterr()
            return ToxRunOutcome(args, self.path, code, out, err, state)

    def __repr__(self) -> str:
        return f"{type(self).__name__}(path={self.path}) at {id(self)}"


@pytest.fixture(autouse=True, scope="session")
def enable_pep517_backend_coverage() -> Iterator[None]:
    try:
        import coverage  # noqa: F401
    except ImportError:  # pragma: no cover
        yield  # pragma: no cover
        return  # pragma: no cover
    # the COV_ env variables needs to be passed on for the PEP-517 backend
    from tox.tox_env.python.virtual_env.package.api import Pep517VirtualEnvPackage

    def default_pass_env(self: Pep517VirtualEnvPackage) -> List[str]:
        result = previous(self)
        result.append("COV_*")
        return result

    previous = Pep517VirtualEnvPackage.default_pass_env
    try:
        Pep517VirtualEnvPackage.default_pass_env = default_pass_env  # type: ignore
        yield
    finally:
        Pep517VirtualEnvPackage.default_pass_env = previous  # type: ignore


class ToxRunOutcome:
    def __init__(self, cmd: Sequence[str], cwd: Path, code: int, out: str, err: str, state: Optional[State]) -> None:
        extended_cmd = [sys.executable, "-m", "tox"]
        extended_cmd.extend(cmd)
        self.cmd: List[str] = extended_cmd
        self.cwd: Path = cwd
        self.code: int = code
        self.out: str = out
        self.err: str = err
        self._state: Optional[State] = state

    @property
    def state(self) -> State:
        if self._state is None:
            raise RuntimeError("no state")
        return self._state

    @property
    def success(self) -> bool:
        return self.code == Outcome.OK

    def assert_success(self) -> None:
        assert self.success, repr(self)

    def assert_failed(self, code: Optional[int] = None) -> None:
        status_match = self.code != 0 if code is None else self.code == code
        assert status_match, f"should be {code}, got {self}"

    def __repr__(self) -> str:
        return "\n".join(
            "{}{}{}".format(k, "\n" if "\n" in v else ": ", v)
            for k, v in (
                ("code", str(self.code)),
                ("cmd", self.shell_cmd),
                ("cwd", str(self.cwd)),
                ("standard output", self.out),
                ("standard error", self.err),
            )
            if v
        )

    @property
    def shell_cmd(self) -> str:
        return shell_cmd(self.cmd)

    def assert_out_err(self, out: str, err: str, *, dedent: bool = True, regex: bool = False) -> None:
        if dedent:
            out = textwrap.dedent(out).lstrip()
        if regex:
            self.matches(out, self.out, re.MULTILINE)
        else:
            assert self.out == out
        if dedent:
            err = textwrap.dedent(err).lstrip()
        if regex:
            self.matches(err, self.err, re.MULTILINE)
        else:
            assert self.err == err

    @staticmethod
    def matches(pattern: str, text: str, flags: int = 0) -> None:
        try:
            from re_assert import Matches
        except ImportError:  # pragma: no cover # hard to test
            match = re.match(pattern, text, flags)
            if match is None:
                warnings.warn("install the re-assert PyPi package for bette error message", UserWarning)
            assert match
        else:
            assert Matches(pattern, flags=flags) == text


class ToxProjectCreator(Protocol):
    def __call__(self, files: Dict[str, Any], base: Optional[Path] = None) -> ToxProject:
        ...


@pytest.fixture(name="tox_project")
def init_fixture(
    tmp_path: Path, capfd: CaptureFixture, monkeypatch: MonkeyPatch, mocker: MockerFixture
) -> ToxProjectCreator:
    def _init(files: Dict[str, Any], base: Optional[Path] = None) -> ToxProject:
        """create tox  projects"""
        return ToxProject(files, base, tmp_path / "p", capfd, monkeypatch, mocker)

    return _init  # noqa


@pytest.fixture()
def empty_project(tox_project: ToxProjectCreator, monkeypatch: MonkeyPatch) -> ToxProject:
    project = tox_project({"tox.ini": ""})
    monkeypatch.chdir(project.path)
    return project


def pytest_addoption(parser: Parser) -> None:
    parser.addoption("--run-integration", action="store_true", help="run the integration tests")


def pytest_configure(config: PyTestConfig) -> None:
    config.addinivalue_line("markers", "integration")


@pytest.mark.trylast
def pytest_collection_modifyitems(config: PyTestConfig, items: List[Function]) -> None:
    # do not require flags if called directly
    if len(items) == 1:  # pragma: no cover # hard to test
        return

    skip_int = pytest.mark.skip(reason="integration tests not run (no --run-int flag)")

    def is_integration(test_item: Function) -> bool:
        return test_item.get_closest_marker("integration") is not None

    integration_enabled = config.getoption("--run-integration")
    if not integration_enabled:  # pragma: no cover # hard to test
        for item in items:
            if is_integration(item):
                item.add_marker(skip_int)
    # run integration tests after unit tests
    items.sort(key=lambda i: 1 if is_integration(i) else 0)


class Index:
    def __init__(self, base_url: str, name: str, client_cmd_base: List[str]) -> None:
        self._client_cmd_base = client_cmd_base
        self._server_url = base_url
        self.name = name

    @property
    def url(self) -> str:
        return f"{self._server_url}/{self.name}/+simple"

    def upload(self, files: Sequence[Path]) -> None:
        check_call(self._client_cmd_base + ["upload", "--index", self.name] + [str(i) for i in files])

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(url={self.url})"  # pragma: no cover

    def use(self, monkeypatch: MonkeyPatch) -> None:
        enable_pypi_server(monkeypatch, self.url)


def enable_pypi_server(monkeypatch: MonkeyPatch, url: Optional[str]) -> None:
    if url is None:  # pragma: no cover # only one of the branches can be hit depending on env
        monkeypatch.delenv("PIP_INDEX_URL", raising=False)
    else:  # pragma: no cover
        monkeypatch.setenv("PIP_INDEX_URL", url)
    monkeypatch.setenv("PIP_RETRIES", str(5))
    monkeypatch.setenv("PIP_TIMEOUT", str(2))


def _find_free_port() -> int:
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as socket_handler:
        socket_handler.bind(("", 0))
        return cast(int, socket_handler.getsockname()[1])


class IndexServer:
    def __init__(self, path: Path) -> None:
        self.path = path

        self.host, self.port = "localhost", _find_free_port()
        self._passwd = "".join(random.choice(string.ascii_letters) for _ in range(8))

        def _exe(name: str) -> str:
            return str(Path(scripts_dir) / f"{name}{'.exe' if IS_WIN else ''}")

        scripts_dir = PythonInfo.current().sysconfig_path("scripts")
        self._init: str = _exe("devpi-init")
        self._server: str = _exe("devpi-server")
        self._client: str = _exe("devpi")

        self._server_dir = self.path / "server"
        self._client_dir = self.path / "client"
        self._indexes: Dict[str, Index] = {}
        self._process: Optional["Popen[str]"] = None
        self._has_use = False
        self._stdout_drain: Optional[Thread] = None

    def __enter__(self) -> "IndexServer":
        self._create_and_start_server()
        self._setup_client()
        return self

    def _create_and_start_server(self) -> None:
        self._server_dir.mkdir(exist_ok=True)
        server_at = str(self._server_dir)
        # 1. create the server
        cmd = [self._init, "--serverdir", server_at]
        cmd.extend(("--no-root-pypi", "--role", "standalone", "--root-passwd", self._passwd))
        check_call(cmd, stdout=PIPE, stderr=PIPE)
        # 2. start the server
        cmd = [self._server, "--serverdir", server_at, "--port", str(self.port), "--offline-mode"]
        self._process = Popen(cmd, stdout=PIPE, universal_newlines=True)
        stdout = self._drain_stdout()
        for line in stdout:  # pragma: no branch # will always loop at least once
            if "serving at url" in line:

                def _keep_draining() -> None:
                    for _ in stdout:
                        pass

                # important to keep draining the stdout, otherwise once the buffer is full Windows blocks the processg s
                self._stdout_drain = Thread(target=_keep_draining, name="tox-test-stdout-drain")
                self._stdout_drain.start()
                break

    def _drain_stdout(self) -> Iterator[str]:
        process = cast("Popen[str]", self._process)
        stdout = cast(IO[str], process.stdout)
        while True:
            if process.poll() is not None:  # pragma: no cover
                print(f"devpi server with pid {process.pid} at {self._server_dir} died")
                break
            yield stdout.readline()

    def _setup_client(self) -> None:
        """create a user on the server and authenticate it"""
        self._client_dir.mkdir(exist_ok=True)
        base = ["--clientdir", str(self._client_dir)]
        check_call([self._client, "use"] + base + [self.url], stdout=PIPE, stderr=PIPE)
        check_call([self._client, "login"] + base + ["root", "--password", self._passwd], stdout=PIPE, stderr=PIPE)

    def create_index(self, name: str, *args: str) -> Index:
        if name in self._indexes:  # pragma: no cover
            raise ValueError(f"index {name} already exists")
        base = [self._client, "--clientdir", str(self._client_dir)]
        check_call(base + ["index", "-c", name, *args], stdout=PIPE, stderr=PIPE)
        index = Index(f"{self.url}/root", name, base)
        if not self._has_use:
            self._has_use = True
            check_call(base + ["use", f"root/{name}"], stdout=PIPE, stderr=PIPE)
        self._indexes[name] = index
        return index

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        if self._process is not None:  # pragma: no cover # defend against devpi startup fail
            self._process.terminate()
        if self._stdout_drain is not None and self._stdout_drain.is_alive():  # pragma: no cover # devpi startup fail
            self._stdout_drain.join()

    @property
    def url(self) -> str:
        return f"http://{self.host}:{self.port}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(url={self.url}, indexes={list(self._indexes)})"  # pragma: no cover


@pytest.fixture(scope="session")
def pypi_server(tmp_path_factory: TempPathFactory) -> Iterator[IndexServer]:
    # takes around 2.5s
    path = tmp_path_factory.mktemp("pypi")
    with IndexServer(path) as server:
        server.create_index("empty", "volatile=False")
        yield server


@pytest.fixture(scope="session")
def _invalid_index_fake_port() -> int:
    return _find_free_port()


@pytest.fixture(autouse=True)
def disable_pip_pypi_access(_invalid_index_fake_port: int, monkeypatch: MonkeyPatch) -> Tuple[str, Optional[str]]:
    """set a fake pip index url, tests that want to use a pypi server should create and overwrite this"""
    previous_url = os.environ.get("PIP_INDEX_URL")
    new_url = f"http://localhost:{_invalid_index_fake_port}/bad-pypi-server"
    monkeypatch.setenv("PIP_INDEX_URL", new_url)
    monkeypatch.setenv("PIP_RETRIES", str(0))
    monkeypatch.setenv("PIP_TIMEOUT", str(0.001))
    return new_url, previous_url


@pytest.fixture(name="enable_pip_pypi_access")
def enable_pip_pypi_access_fixture(
    disable_pip_pypi_access: Tuple[str, Optional[str]], monkeypatch: MonkeyPatch
) -> Optional[str]:
    """set a fake pip index url, tests that want to use a pypi server should create and overwrite this"""
    _, previous_url = disable_pip_pypi_access
    enable_pypi_server(monkeypatch, previous_url)
    return previous_url


__all__ = (
    "CaptureFixture",
    "LogCaptureFixture",
    "TempPathFactory",
    "MonkeyPatch",
    "ToxRunOutcome",
    "ToxProject",
    "ToxProjectCreator",
    "check_os_environ",
    "IndexServer",
    "Index",
)



import os
import sys
from abc import ABC, abstractmethod
from contextlib import contextmanager
from enum import Enum
from pathlib import Path
from threading import RLock
from typing import Any, Dict, Iterator, List, NoReturn, Optional, Sequence, Set, Tuple, Union, cast

from packaging.markers import Variable
from packaging.requirements import Requirement

from tox.config.cli.parser import Parsed
from tox.config.sets import ConfigSet
from tox.execute.api import ExecuteStatus
from tox.execute.pep517_backend import LocalSubProcessPep517Executor
from tox.execute.request import StdinSource
from tox.journal import EnvJournal
from tox.report import ToxHandler
from tox.tox_env.errors import Fail
from tox.tox_env.python.package import PythonPackage
from tox.util.pep517.frontend import BackendFailed, CmdStatus, ConfigSettings, Frontend, WheelResult

from ..api import VirtualEnv

if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
    from importlib.metadata import Distribution, PathDistribution  # type: ignore[attr-defined]
else:  # pragma: no cover (<py38)
    from importlib_metadata import Distribution, PathDistribution  # noqa


TOX_PACKAGE_ENV_ID = "virtualenv-pep-517"


class PackageType(Enum):
    sdist = 1
    wheel = 2
    dev = 3
    skip = 4


class ToxBackendFailed(Fail, BackendFailed):
    def __init__(self, backend_failed: BackendFailed) -> None:
        Fail.__init__(self)
        result: Dict[str, Any] = {
            "code": backend_failed.code,
            "exc_type": backend_failed.exc_type,
            "exc_msg": backend_failed.exc_msg,
        }
        BackendFailed.__init__(
            self,
            result,
            backend_failed.out,
            backend_failed.err,
        )


class ToxCmdStatus(CmdStatus):
    def __init__(self, execute_status: ExecuteStatus) -> None:
        self._execute_status = execute_status

    @property
    def done(self) -> bool:
        # 1. process died
        status = self._execute_status
        if status.exit_code is not None:  # pragma: no branch
            return True  # pragma: no cover
        # 2. the backend output reported back that our command is done
        content = status.out
        at = content.rfind(b"Backend: Wrote response ")
        if at != -1 and content.find(b"\n", at) != -1:
            return True
        return False

    def out_err(self) -> Tuple[str, str]:
        status = self._execute_status
        if status is None or status.outcome is None:  # interrupt before status create # pragma: no branch
            return "", ""  # pragma: no cover
        return status.outcome.out_err()


class Pep517VirtualEnvPackage(VirtualEnv, PythonPackage, Frontend, ABC):
    """local file system python virtual environment via the virtualenv package"""

    def __init__(
        self, conf: ConfigSet, core: ConfigSet, options: Parsed, journal: EnvJournal, log_handler: ToxHandler
    ) -> None:
        VirtualEnv.__init__(self, conf, core, options, journal, log_handler)
        Frontend.__init__(self, *Frontend.create_args_from_folder(core["tox_root"]))
        self._distribution_meta: Optional[PathDistribution] = None  # type: ignore[no-any-unimported]
        self._build_requires: Optional[Tuple[Requirement]] = None
        self._build_wheel_cache: Optional[WheelResult] = None
        self._backend_executor: Optional[LocalSubProcessPep517Executor] = None
        self._package_dependencies: Optional[List[Requirement]] = None
        self._lock = RLock()  # can build only one package at a time
        self._package: Optional[Path] = None
        self._teardown_done = False

    def register_config(self) -> None:
        super().register_config()
        self.conf.add_config(
            keys=["meta_dir"],
            of_type=Path,
            default=lambda conf, name: cast(Path, self.conf["env_dir"]) / ".meta",
            desc="directory assigned to the tox environment",
        )
        self.conf.add_config(
            keys=["pkg_dir"],
            of_type=Path,
            default=lambda conf, name: cast(Path, self.conf["env_dir"]) / "dist",
            desc="directory assigned to the tox environment",
        )

    @property
    def meta_folder(self) -> Path:
        meta_folder: Path = self.conf["meta_dir"]
        meta_folder.mkdir(exist_ok=True)
        return meta_folder

    def _ensure_meta_present(self) -> None:
        if self._distribution_meta is not None:  # pragma: no branch
            return  # pragma: no cover
        self.ensure_setup()
        dist_info = self.prepare_metadata_for_build_wheel(self.meta_folder).metadata
        self._distribution_meta = Distribution.at(str(dist_info))

    @abstractmethod
    def _build_artifact(self) -> Path:
        raise NotImplementedError

    def perform_packaging(self) -> List[Path]:
        """build_wheel/build_sdist"""
        with self._lock:
            if self._package is None:
                self.ensure_setup()
                self._package = self._build_artifact()
        return [self._package]

    def get_package_dependencies(self, extras: Set[str]) -> List[Requirement]:
        with self._lock:
            if self._package_dependencies is None:  # pragma: no branch
                self._ensure_meta_present()
                self._package_dependencies = self.discover_package_dependencies(self._distribution_meta, extras)
        return self._package_dependencies

    @staticmethod
    def discover_package_dependencies(  # type: ignore[no-any-unimported]
        meta: PathDistribution, extras: Set[str]
    ) -> List[Requirement]:
        result: List[Requirement] = []
        requires = meta.requires or []
        for req_str in requires:
            req = Requirement(req_str)
            markers: List[Union[str, Tuple[Variable, Variable, Variable]]] = getattr(req.marker, "_markers", []) or []

            # find the extra marker (if has)
            _at: Optional[int] = None
            extra: Optional[str] = None
            for _at, (marker_key, op, marker_value) in (
                (_at_marker, marker)
                for _at_marker, marker in enumerate(markers)
                if isinstance(marker, tuple) and len(marker) == 3
            ):
                if marker_key.value == "extra" and op.value == "==":  # pragma: no branch
                    extra = marker_value.value
                    break
            # continue only if this extra should be included
            if not (extra is None or extra in extras):
                continue
            # delete the extra marker if present
            if _at is not None:
                del markers[_at]
                _at -= 1
                if _at > 0 and (isinstance(markers[_at], str) and markers[_at] in ("and", "or")):
                    del markers[_at]
                if len(markers) == 0:
                    req.marker = None
            result.append(req)
        return result

    @property
    def backend_executor(self) -> LocalSubProcessPep517Executor:
        if self._backend_executor is None:
            self._backend_executor = LocalSubProcessPep517Executor(
                colored=self.options.is_colored,
                cmd=self.backend_cmd,
                env=self.environment_variables,
                cwd=self._root,
            )

        return self._backend_executor

    @property
    def pkg_dir(self) -> Path:
        return cast(Path, self.conf["pkg_dir"])

    @property
    def backend_cmd(self) -> Sequence[str]:
        return ["python"] + self.backend_args

    @property
    def environment_variables(self) -> Dict[str, str]:
        env = super().environment_variables
        env["PYTHONPATH"] = os.pathsep.join(str(i) for i in self._backend_paths)
        return env

    def teardown(self) -> None:
        self.ref_count.decrement()
        if self.ref_count.value == 0 and self._backend_executor is not None and self._teardown_done is False:
            self._teardown_done = True
            try:
                if self.backend_executor.is_alive:
                    self._send("_exit")  # try first on amicable shutdown
            except SystemExit:  # if already has been interrupted ignore
                pass
            finally:
                self._backend_executor.close()

    @contextmanager
    def _send_msg(self, cmd: str, result_file: Path, msg: str) -> Iterator[ToxCmdStatus]:  # type: ignore[override]
        with self.execute_async(
            cmd=self.backend_cmd,
            cwd=self._root,
            stdin=StdinSource.API,
            show=None,
            run_id=cmd,
            executor=self.backend_executor,
        ) as execute_status:
            execute_status.write_stdin(f"{msg}{os.linesep}")
            yield ToxCmdStatus(execute_status)
        outcome = execute_status.outcome
        if outcome is not None:  # pragma: no branch
            outcome.assert_success()

    @contextmanager
    def _wheel_directory(self) -> Iterator[Path]:
        yield self.pkg_dir  # use our local wheel directory

    def build_wheel(
        self,
        wheel_directory: Path,
        config_settings: Optional[ConfigSettings] = None,
        metadata_directory: Optional[Path] = None,
    ) -> WheelResult:
        # only build once a wheel per session - might need more than once if the backend doesn't
        # support prepare metadata for wheel
        if self._build_wheel_cache is None:
            self._build_wheel_cache = super().build_wheel(wheel_directory, config_settings, metadata_directory)
        return self._build_wheel_cache

    def _send(self, cmd: str, **kwargs: Any) -> Tuple[Any, str, str]:
        try:
            return super()._send(cmd, **kwargs)
        except BackendFailed as exception:
            raise exception if isinstance(exception, ToxBackendFailed) else ToxBackendFailed(exception) from exception

    def _unexpected_response(self, cmd: str, got: Any, expected_type: Any, out: str, err: str) -> NoReturn:
        try:
            super()._unexpected_response(cmd, got, expected_type, out, err)
        except BackendFailed as exception:
            raise exception if isinstance(exception, ToxBackendFailed) else ToxBackendFailed(exception) from exception

    def requires(self) -> Tuple[Requirement, ...]:
        return self._requires



from pathlib import Path

from tox.config.cli.parser import ToxParser
from tox.config.loader.memory import MemoryLoader
from tox.plugin.impl import impl
from tox.report import HandledError
from tox.session.cmd.run.common import env_run_create_flags
from tox.session.cmd.run.sequential import run_sequential
from tox.session.common import CliEnv, env_list_flag
from tox.session.state import State


@impl
def tox_add_option(parser: ToxParser) -> None:
    our = parser.add_command(
        "devenv",
        ["d"],
        "sets up a development environment at ENVDIR based on the env's tox configuration specified ",
        devenv,
    )
    our.add_argument("devenv_path", metavar="path", default=Path("venv").absolute(), nargs="?")
    env_list_flag(our, default=CliEnv("py"))
    env_run_create_flags(our)


def devenv(state: State) -> int:
    env_list = list(state.env_list(everything=False))
    if len(env_list) != 1:
        raise HandledError(f"exactly one target environment allowed in devenv mode but found {', '.join(env_list)}")
    loader = MemoryLoader(  # these configuration values are loaded from in-memory always (no file conf)
        # dev environments must be of type dev
        usedevelop=True,
        # move it in source
        env_dir=Path(state.options.devenv_path),
    )

    state.options.no_test = True  # do not run the test phase
    state.conf.get_env(env_list[0], loaders=[loader])
    return run_sequential(state)



from pathlib import Path
from typing import Any, Dict, Iterator, Optional, Set, Tuple, Type, cast

from tox.config.loader.convert import T
from tox.config.main import Config
from tox.config.types import Command, EnvList

from .api import Loader


class MemoryLoader(Loader[Any]):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__([])
        self.raw: Dict[str, Any] = {**kwargs}

    def load_raw(self, key: Any, conf: Optional["Config"], env_name: Optional[str]) -> T:
        return cast(T, self.raw[key])

    def found_keys(self) -> Set[str]:
        return set(self.raw.keys())

    @staticmethod
    def to_bool(value: Any) -> bool:
        return value  # type: ignore[no-any-return]

    @staticmethod
    def to_str(value: Any) -> str:
        return value  # type: ignore[no-any-return]

    @staticmethod
    def to_list(value: Any, of_type: Type[Any]) -> Iterator[T]:
        return value  # type: ignore[no-any-return]

    @staticmethod
    def to_set(value: Any, of_type: Type[Any]) -> Iterator[T]:
        return iter(value)  # type: ignore[no-any-return]

    @staticmethod
    def to_dict(value: Any, of_type: Tuple[Type[Any], Type[Any]]) -> Iterator[Tuple[T, T]]:
        return value.items()  # type: ignore[no-any-return]

    @staticmethod
    def to_path(value: Any) -> Path:
        return value  # type: ignore[no-any-return]

    @staticmethod
    def to_command(value: Any) -> Command:
        return value  # type: ignore[no-any-return]

    @staticmethod
    def to_env_list(value: Any) -> EnvList:
        return value  # type: ignore[no-any-return]



"""
This package handles provisioning an appropriate tox version per requirements.
"""
import logging
import sys
from argparse import ArgumentParser
from typing import List, Tuple, Union, cast

from packaging.requirements import Requirement
from packaging.utils import canonicalize_name
from packaging.version import Version

from tox.config.loader.memory import MemoryLoader
from tox.config.main import Config
from tox.config.sets import ConfigSet
from tox.execute.api import StdinSource
from tox.plugin.impl import impl
from tox.session.state import State
from tox.tox_env.python.req_file import RequirementsFile
from tox.tox_env.python.runner import PythonRun
from tox.tox_env.python.virtual_env.package.api import PackageType
from tox.version import __version__ as current_version

if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
    from importlib.metadata import PackageNotFoundError, distribution
else:  # pragma: no cover (py38+)
    from importlib_metadata import PackageNotFoundError, distribution  # noqa


@impl
def tox_add_option(parser: ArgumentParser) -> None:
    parser.add_argument(
        "--no-recreate-provision",
        dest="recreate",
        help="if recreate is set do not recreate provision tox environment",
        action="store_true",
    )


@impl
def tox_add_core_config(core: ConfigSet) -> None:
    core.add_config(
        keys=["min_version", "minversion"],
        of_type=Version,
        # do not include local version specifier (because it's not allowed in version spec per PEP-440)
        default=Version(current_version.split("+")[0]),
        desc="Define the minimal tox version required to run",
    )
    core.add_config(
        keys="provision_tox_env",
        of_type=str,
        default=".tox",
        desc="Name of the virtual environment used to provision a tox.",
    )

    def add_tox_requires_min_version(requires: List[Requirement], conf: Config) -> List[Requirement]:
        min_version: Version = conf.core["min_version"]
        requires.append(Requirement(f"tox >= {min_version.public}"))
        return requires

    core.add_config(
        keys="requires",
        of_type=List[Requirement],
        default=[],
        desc="Name of the virtual environment used to provision a tox.",
        post_process=add_tox_requires_min_version,
    )


def provision(state: State) -> Union[int, bool]:
    requires: List[Requirement] = state.conf.core["requires"]
    missing: List[Tuple[Requirement, str]] = []
    for package in requires:
        package_name = canonicalize_name(package.name)
        try:
            dist = distribution(package_name)
            if not package.specifier.contains(dist.version, prereleases=True):
                missing.append((package, dist.version))
        except PackageNotFoundError:
            missing.append((package, "N/A"))
    if not missing:
        return False
    deps = ", ".join(f"{p} ({ver})" for p, ver in missing)
    logging.warning(
        "will run in automatically provisioned tox, host %s is missing [requires (has)]: %s", sys.executable, deps
    )
    return run_provision(requires, state)


def run_provision(deps: List[Requirement], state: State) -> int:  # noqa
    """"""
    loader = MemoryLoader(  # these configuration values are loaded from in-memory always (no file conf)
        base=[],  # disable inheritance for provision environments
        package=PackageType.skip,  # no packaging for this please
        # use our own dependency specification
        deps=RequirementsFile("\n".join(str(d) for d in deps), root=state.conf.core["tox_root"]),
        pass_env=["*"],  # do not filter environment variables, will be handled by provisioned tox
    )
    provision_tox_env: str = state.conf.core["provision_tox_env"]
    state.conf.get_env(provision_tox_env, loaders=[loader])
    tox_env = cast(PythonRun, state.tox_env(provision_tox_env))
    env_python = tox_env.env_python()
    logging.info("will run in a automatically provisioned python environment under %s", env_python)
    recreate = state.options.no_recreate_provision is False if state.options.recreate else False
    tox_env.ensure_setup(recreate=recreate)
    args: List[str] = [str(env_python), "-m", "tox"]
    args.extend(state.args)
    outcome = tox_env.execute(cmd=args, stdin=StdinSource.user_only(), show=True, run_id="provision")
    return cast(int, outcome.exit_code)



"""
Group together configuration values that belong together (such as base tox configuration, tox environment configs)
"""
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Generic, Iterable, List, Optional, Type, TypeVar, Union, cast

from tox.config.loader.api import Loader

if TYPE_CHECKING:
    from tox.config.main import Config  # pragma: no cover


T = TypeVar("T")
V = TypeVar("V")


class ConfigDefinition(ABC, Generic[T]):
    """Abstract base class for configuration definitions"""

    def __init__(self, keys: Iterable[str], desc: str, env_name: Optional[str]) -> None:
        self.keys = keys
        self.desc = desc
        self.env_name = env_name

    @abstractmethod
    def __call__(self, conf: "Config", key: Optional[str], loaders: List[Loader[T]]) -> T:
        raise NotImplementedError

    def __eq__(self, o: Any) -> bool:
        return type(self) == type(o) and (self.keys, self.desc, self.env_name) == (o.keys, o.desc, o.env_name)

    def __ne__(self, o: Any) -> bool:
        return not (self == o)


class ConfigConstantDefinition(ConfigDefinition[T]):
    """A configuration definition whose value is defined upfront (such as the tox environment name)"""

    def __init__(
        self,
        keys: Iterable[str],
        desc: str,
        env_name: Optional[str],
        value: Union[Callable[[], T], T],
    ) -> None:
        super().__init__(keys, desc, env_name)
        self.value = value

    def __call__(self, conf: "Config", name: Optional[str], loaders: List[Loader[T]]) -> T:
        if callable(self.value):
            value = self.value()
        else:
            value = self.value
        return value

    def __eq__(self, o: Any) -> bool:
        return type(self) == type(o) and super().__eq__(o) and self.value == o.value


_PLACE_HOLDER = object()


class ConfigDynamicDefinition(ConfigDefinition[T]):
    """A configuration definition that comes from a source (such as in memory, an ini file, a toml file, etc.)"""

    def __init__(
        self,
        keys: Iterable[str],
        desc: str,
        env_name: Optional[str],
        of_type: Type[T],
        default: Union[Callable[["Config", Optional[str]], T], T],
        post_process: Optional[Callable[[T, "Config"], T]] = None,
    ) -> None:
        super().__init__(keys, desc, env_name)
        self.of_type = of_type
        self.default = default
        self.post_process = post_process
        self._cache: Union[object, T] = _PLACE_HOLDER

    def __call__(self, conf: "Config", name: Optional[str], loaders: List[Loader[T]]) -> T:
        if self._cache is _PLACE_HOLDER:
            found = False
            for key in self.keys:
                for loader in loaders:
                    try:
                        value = loader.load(key, self.of_type, conf, self.env_name)
                        found = True
                    except KeyError:
                        continue
                    break
                if found:
                    break
            else:
                value = self.default(conf, self.env_name) if callable(self.default) else self.default
            if self.post_process is not None:
                value = self.post_process(value, conf)  # noqa
            self._cache = value
        return cast(T, self._cache)

    def __repr__(self) -> str:
        values = ((k, v) for k, v in vars(self).items() if k != "post_process" and v is not None)
        return f"{type(self).__name__}({', '.join('{}={}'.format(k, v) for k,v in values)})"

    def __eq__(self, o: Any) -> bool:
        return (
            type(self) == type(o)
            and super().__eq__(o)
            and (self.of_type, self.default, self.post_process) == (o.of_type, o.default, o.post_process)
        )



from tox.config.cli.parser import ToxParser
from tox.plugin.impl import impl
from tox.session.state import State


@impl
def tox_add_option(parser: ToxParser) -> None:
    parser.add_command(
        "quickstart",
        ["q"],
        "Command-line script to quickly tox config file for a Python project",
        quickstart,
    )


def quickstart(state: State) -> int:
    print("done quickstart")
    return 0



from pathlib import Path
from typing import Any, Tuple

from packaging.requirements import Requirement

from tox.plugin.impl import impl
from tox.tox_env.register import ToxEnvRegister
from tox.util.pep517.frontend import BackendFailed

from ..api import Pep517VirtualEnvPackage


class Pep517VirtualEnvPackageWheel(Pep517VirtualEnvPackage):
    @staticmethod
    def id() -> str:
        return "virtualenv-pep-517-wheel"

    def build_requires(self) -> Tuple[Requirement, ...]:
        result = self.get_requires_for_build_wheel()
        return result.requires

    def _build_artifact(self) -> Path:
        result = self.build_wheel(
            wheel_directory=self.pkg_dir,
            metadata_directory=self.meta_folder,
            config_settings={"--global-option": ["--bdist-dir", str(self.conf["env_dir"] / "build")]},
        )
        return result.wheel

    def _send(self, cmd: str, **kwargs: Any) -> Tuple[Any, str, str]:
        if cmd == "prepare_metadata_for_build_wheel":
            # given we'll build a wheel we might skip the prepare step
            result = {
                "code": 1,
                "exc_type": "AvoidRedundant",
                "exc_msg": "will need to build wheel either way, avoid prepare",
            }
            raise BackendFailed(result, "", "")
        return super()._send(cmd, **kwargs)


@impl
def tox_register_tox_env(register: ToxEnvRegister) -> None:
    register.add_package_env(Pep517VirtualEnvPackageWheel)



import pytest

from tox.config.cli.parse import get_options
from tox.config.loader.api import Override
from tox.pytest import CaptureFixture


@pytest.mark.parametrize("flag", ["-x", "--override"])
def test_override_incorrect(flag: str, capsys: CaptureFixture) -> None:
    with pytest.raises(SystemExit):
        get_options(flag, "magic")
    out, err = capsys.readouterr()
    assert not out
    assert "override magic has no = sign in it" in err


@pytest.mark.parametrize("flag", ["-x", "--override"])
def test_override_add(flag: str) -> None:
    parsed, _, __, ___ = get_options(flag, "magic=true")
    assert len(parsed.override) == 1
    value = parsed.override[0]
    assert value.key == "magic"
    assert value.value == "true"
    assert value.namespace == ""


def test_override_equals() -> None:
    assert Override("a=b") == Override("a=b")


def test_override_not_equals() -> None:
    assert Override("a=b") != Override("c=d")


def test_override_not_equals_different_type() -> None:
    assert Override("a=b") != 1


def test_override_repr() -> None:
    assert repr(Override("b.a=c")) == "Override('b.a=c')"



"""Hook specifications for the tox project - see https://pluggy.readthedocs.io/"""
from argparse import ArgumentParser
from typing import Any, Callable, Type, TypeVar, cast

import pluggy

from tox.config.sets import ConfigSet
from tox.tox_env.api import ToxEnv
from tox.tox_env.register import ToxEnvRegister

from . import NAME

F = TypeVar("F", bound=Callable[..., Any])
_hook_spec = pluggy.HookspecMarker(NAME)


def hook_spec(func: F) -> F:
    return cast(F, _hook_spec(func))


@hook_spec
def tox_add_option(parser: ArgumentParser) -> None:  # noqa
    """add cli flags"""


@hook_spec
def tox_add_core_config(core: ConfigSet) -> None:  # noqa
    """add options to the core section of the tox"""


@hook_spec
def tox_register_tox_env(register: ToxEnvRegister) -> Type[ToxEnv]:  # noqa
    """register new tox environment types that can have their own argument"""



"""
Declare the abstract base class for tox environments that handle the Python language.
"""
import logging
import sys
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Sequence, Union, cast

from packaging.requirements import Requirement
from virtualenv.discovery.py_spec import PythonSpec

from tox.config.cli.parser import Parsed
from tox.config.main import Config
from tox.config.sets import ConfigSet
from tox.journal import EnvJournal
from tox.report import ToxHandler
from tox.tox_env.api import ToxEnv
from tox.tox_env.errors import Fail, Recreate, Skip


class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: str
    serial: int


class PythonInfo(NamedTuple):
    executable: Path
    implementation: str
    version_info: VersionInfo
    version: str
    is_64: bool
    platform: str
    extra_version_info: Optional[str]


class PythonDep:
    def __init__(self, value: Union[Path, Requirement]) -> None:
        self._value = value

    @property
    def value(self) -> Union[Path, Requirement]:
        return self._value

    def __str__(self) -> str:
        return str(self._value)

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and str(self) == str(other)

    def __ne__(self, other: Any) -> bool:
        return not (self == other)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(value={self.value!r})"


PythonDeps = Sequence[PythonDep]


class Python(ToxEnv, ABC):
    def __init__(
        self, conf: ConfigSet, core: ConfigSet, options: Parsed, journal: EnvJournal, log_handler: ToxHandler
    ) -> None:
        self._base_python: Optional[PythonInfo] = None
        self._base_python_searched: bool = False
        super().__init__(conf, core, options, journal, log_handler)

    def register_config(self) -> None:
        super().register_config()
        self.conf.add_config(
            keys=["base_python", "basepython"],
            of_type=List[str],
            default=self.default_base_python,
            desc="environment identifier for python, first one found wins",
        )
        self.conf.add_constant(
            keys=["env_site_packages_dir", "envsitepackagesdir"],
            desc="the python environments site package",
            value=lambda: self.env_site_package_dir(),
        )
        self.conf.add_constant(
            ["env_python", "envpython"],
            desc="python executable from within the tox environment",
            value=lambda: self.env_python(),
        )

    def default_pass_env(self) -> List[str]:
        env = super().default_pass_env()
        if sys.platform == "win32":  # pragma: win32 cover
            env.extend(
                [
                    "SYSTEMROOT",  # needed for python's crypto module
                    "PATHEXT",  # needed for discovering executables
                    "COMSPEC",  # needed for distutils cygwin compiler
                    "PROCESSOR_ARCHITECTURE",  # platform.machine()
                    "USERPROFILE",  # needed for `os.path.expanduser()`
                    "MSYSTEM",  # controls paths printed format
                ]
            )
        return env

    def default_base_python(self, conf: "Config", env_name: Optional[str]) -> List[str]:
        spec = PythonSpec.from_string_spec(env_name)
        if spec.implementation is not None:
            if spec.implementation.lower() in ("cpython", "pypy") and env_name is not None:
                return [env_name]
        return [sys.executable]

    @abstractmethod
    def env_site_package_dir(self) -> Path:
        """
        If we have the python we just need to look at the last path under prefix.
        E.g., Debian derivatives change the site-packages to dist-packages, so we need to fix it for site-packages.
        """
        raise NotImplementedError

    @abstractmethod
    def env_python(self) -> Path:
        """The python executable within the tox environment"""
        raise NotImplementedError

    def setup(self) -> None:
        """setup a virtual python environment"""
        conf = self.python_cache()
        with self._cache.compare(conf, Python.__name__) as (eq, old):
            if eq is False:  # if changed create
                self.create_python_env()
            self._paths = self.paths()
        super().setup()

    def setup_has_been_done(self) -> None:
        """called when setup is done"""
        super().setup_has_been_done()
        if self.journal:
            outcome = self.get_installed_packages()
            self.journal["installed_packages"] = outcome

    @abstractmethod
    def get_installed_packages(self) -> List[str]:
        raise NotImplementedError

    def python_cache(self) -> Dict[str, Any]:
        return {
            "version_info": list(self.base_python.version_info),
            "executable": str(self.base_python.executable),
        }

    @property
    def base_python(self) -> PythonInfo:
        """Resolve base python"""
        if self._base_python_searched is False:
            base_pythons = self.conf["base_python"]
            self._base_python_searched = True
            self._base_python = self._get_python(base_pythons)
            if self._base_python is None:
                if self.core["skip_missing_interpreters"]:
                    raise Skip
                raise NoInterpreter(base_pythons)
            if self.journal:
                value = {
                    "executable": str(self._base_python.executable),
                    "implementation": self._base_python.implementation,
                    "version_info": tuple(self.base_python.version_info),
                    "version": self._base_python.version,
                    "is_64": self._base_python.is_64,
                    "sysplatform": self._base_python.platform,
                    "extra_version_info": None,
                }
                self.journal["python"] = value
        return cast(PythonInfo, self._base_python)

    @abstractmethod
    def _get_python(self, base_python: List[str]) -> Optional[PythonInfo]:
        raise NotImplementedError

    def cached_install(self, deps: PythonDeps, section: str, of_type: str) -> bool:
        conf_deps: List[str] = [str(i) for i in deps]
        with self._cache.compare(conf_deps, section, of_type) as (eq, old):
            if eq is True:
                return True
            if old is None:
                old = []
            missing = [PythonDep(Requirement(i)) for i in (set(old) - set(conf_deps))]
            if missing:  # no way yet to know what to uninstall here (transitive dependencies?)
                # bail out and force recreate
                logging.warning(f"recreate env because dependencies removed: {', '.join(str(i) for i in missing)}")
                raise Recreate
            new_deps_str = set(conf_deps) - set(old)
            new_deps = [PythonDep(Requirement(i)) for i in new_deps_str]
            self.install_python_packages(packages=new_deps, of_type=of_type)
        return False

    @abstractmethod
    def create_python_env(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def paths(self) -> List[Path]:
        raise NotImplementedError

    @abstractmethod
    def install_python_packages(self, packages: PythonDeps, of_type: str, no_deps: bool = False) -> None:
        raise NotImplementedError


class NoInterpreter(Fail):
    """could not find interpreter"""

    def __init__(self, base_pythons: List[str]) -> None:
        self.base_pythons = base_pythons

    def __str__(self) -> str:
        return f"could not find python interpreter matching any of the specs {', '.join(self.base_pythons)}"



from .api import PackageType
from .artifact.dev import LegacyDevVirtualEnvPackage
from .artifact.sdist import Pep517VirtualEnvPackageSdist
from .artifact.wheel import Pep517VirtualEnvPackageWheel


def virtual_env_package_id(of_type: PackageType) -> str:
    if of_type is PackageType.sdist:
        return Pep517VirtualEnvPackageSdist.id()
    elif of_type is PackageType.wheel:
        return Pep517VirtualEnvPackageWheel.id()
    elif of_type is PackageType.dev:  # pragma: no branch
        return LegacyDevVirtualEnvPackage.id()
    raise KeyError(PackageType.name)  # pragma: no cover # can only happen if we introduce new types and forget to set



from collections import OrderedDict
from typing import Dict, Set

import pytest

from tox.util.graph import stable_topological_sort


def test_topological_order_empty() -> None:
    graph: Dict[str, Set[str]] = OrderedDict()
    result = stable_topological_sort(graph)
    assert result == []


def test_topological_order_specified_only() -> None:
    graph: Dict[str, Set[str]] = OrderedDict()
    graph["A"] = {"B", "C"}
    result = stable_topological_sort(graph)
    assert result == ["A"]


def test_topological_order() -> None:
    graph: Dict[str, Set[str]] = OrderedDict()
    graph["A"] = {"B", "C"}
    graph["B"] = set()
    graph["C"] = set()
    result = stable_topological_sort(graph)
    assert result == ["B", "C", "A"]


def test_topological_order_cycle() -> None:
    graph: Dict[str, Set[str]] = OrderedDict()
    graph["A"] = {"B", "C"}
    graph["B"] = {"A"}
    with pytest.raises(ValueError, match="A | B"):
        stable_topological_sort(graph)


def test_topological_complex() -> None:
    graph: Dict[str, Set[str]] = OrderedDict()
    graph["A"] = {"B", "C"}
    graph["B"] = {"C", "D"}
    graph["C"] = {"D"}
    graph["D"] = set()
    result = stable_topological_sort(graph)
    assert result == ["D", "C", "B", "A"]


def test_two_sub_graph() -> None:
    graph: Dict[str, Set[str]] = OrderedDict()
    graph["F"] = set()
    graph["E"] = set()
    graph["D"] = {"E", "F"}
    graph["A"] = {"B", "C"}
    graph["B"] = set()
    graph["C"] = set()

    result = stable_topological_sort(graph)
    assert result == ["F", "E", "D", "B", "C", "A"]


def test_two_sub_graph_circle() -> None:
    graph: Dict[str, Set[str]] = OrderedDict()
    graph["F"] = set()
    graph["E"] = set()
    graph["D"] = {"E", "F"}
    graph["A"] = {"B", "C"}
    graph["B"] = {"A"}
    graph["C"] = set()
    with pytest.raises(ValueError, match="A | B"):
        stable_topological_sort(graph)



"""A minimal non-colored version of https://pypi.org/project/halo, to track list progress"""
import os
import sys
import threading
import time
from collections import OrderedDict
from types import TracebackType
from typing import IO, Dict, Optional, Sequence, Type, TypeVar

from colorama import Fore

if sys.platform == "win32":  # pragma: win32 cover
    import ctypes

    class _CursorInfo(ctypes.Structure):
        _fields_ = [("size", ctypes.c_int), ("visible", ctypes.c_byte)]


def _file_support_encoding(chars: Sequence[str], file: IO[str]) -> bool:
    encoding = getattr(file, "encoding", None)
    if encoding is not None:  # pragma: no branch  # this should be always set, unless someone passes in something bad
        for char in chars:
            try:
                char.encode(encoding)
            except UnicodeEncodeError:
                break
        else:
            return True
    return False


T = TypeVar("T", bound="Spinner")
MISS_DURATION = 0.01


class Spinner:
    CLEAR_LINE = "\033[K"
    max_width = 120
    UNICODE_FRAMES = ["⠋", "⠙", "⠹", "⠸", "⠼", "⠴", "⠦", "⠧", "⠇", "⠏"]
    ASCII_FRAMES = ["|", "-", "+", "x", "*"]

    def __init__(
        self,
        enabled: bool = True,
        refresh_rate: float = 0.1,
        colored: bool = True,
        stream: Optional[IO[str]] = None,
        total: Optional[int] = None,
    ) -> None:
        self.is_colored = colored
        self.refresh_rate = refresh_rate
        self.enabled = enabled
        stream = sys.stdout if stream is None else stream
        self.frames = self.UNICODE_FRAMES if _file_support_encoding(self.UNICODE_FRAMES, stream) else self.ASCII_FRAMES
        self.stream = stream
        self.total = total

        self._envs: Dict[str, float] = OrderedDict()
        self._frame_index = 0

    def clear(self) -> None:
        if self.enabled:
            self.stream.write("\r")
            self.stream.write(self.CLEAR_LINE)

    def render(self) -> "Spinner":
        while True:
            self._stop_spinner.wait(self.refresh_rate)
            if self._stop_spinner.is_set():
                break
            self.render_frame()
        return self

    def render_frame(self) -> None:
        if self.enabled:
            self.clear()
            self.stream.write(f"\r{self.frame()}")

    def frame(self) -> str:
        frame = self.frames[self._frame_index]
        self._frame_index += 1
        self._frame_index %= len(self.frames)
        total = f"/{self.total}" if self.total is not None else ""
        text_frame = f"[{len(self._envs)}{total}] {' | '.join(self._envs)}"
        if len(text_frame) > self.max_width - 1:
            text_frame = "{}...".format(text_frame[: self.max_width - 1 - 3])
        return "{} {}".format(*[(frame, text_frame)][0])

    def __enter__(self: T) -> T:
        if self.enabled:
            self.disable_cursor()
        self.render_frame()
        self._stop_spinner = threading.Event()
        self._spinner_thread = threading.Thread(target=self.render)
        self._spinner_thread.setDaemon(True)
        self._spinner_thread.start()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        if not self._stop_spinner.is_set():  # pragma: no branch
            if self._spinner_thread:  # pragma: no branch # hard to test
                self._stop_spinner.set()
                self._spinner_thread.join()

            self._frame_index = 0
            if self.enabled:
                self.clear()
                self.enable_cursor()

    def add(self, name: str) -> None:
        self._envs[name] = time.monotonic()

    def succeed(self, key: str) -> None:
        self.finalize(key, "OK ✔", Fore.GREEN)

    def fail(self, key: str) -> None:
        self.finalize(key, "FAIL ✖", Fore.RED)

    def skip(self, key: str) -> None:
        self.finalize(key, "SKIP ⚠", Fore.YELLOW)

    def finalize(self, key: str, status: str, color: int) -> None:
        start_at = self._envs.pop(key, None)
        if self.enabled:
            self.clear()
        duration = MISS_DURATION if start_at is None else time.monotonic() - start_at
        base = f"{key}: {status} in {td_human_readable(duration)}"
        if self.is_colored:
            base = f"{color}{base}{Fore.RESET}"
        base += os.linesep
        self.stream.write(base)

    def disable_cursor(self) -> None:
        if self.stream.isatty():
            if sys.platform == "win32":  # pragma: win32 cover
                ci = _CursorInfo()
                handle = ctypes.windll.kernel32.GetStdHandle(-11)
                ctypes.windll.kernel32.GetConsoleCursorInfo(handle, ctypes.byref(ci))
                ci.visible = False
                ctypes.windll.kernel32.SetConsoleCursorInfo(handle, ctypes.byref(ci))
            else:
                self.stream.write("\033[?25l")

    def enable_cursor(self) -> None:
        if self.stream.isatty():
            if sys.platform == "win32":  # pragma: win32 cover
                ci = _CursorInfo()
                handle = ctypes.windll.kernel32.GetStdHandle(-11)
                ctypes.windll.kernel32.GetConsoleCursorInfo(handle, ctypes.byref(ci))
                ci.visible = True
                ctypes.windll.kernel32.SetConsoleCursorInfo(handle, ctypes.byref(ci))
            else:
                self.stream.write("\033[?25h")


_PERIODS = [
    ("day", 60 * 60 * 24),
    ("hour", 60 * 60),
    ("minute", 60),
    ("second", 1),
]


def td_human_readable(seconds: float) -> str:
    texts = []
    total_seconds = seconds
    for period_name, period_seconds in _PERIODS:
        if seconds > period_seconds or period_seconds == 1:
            period_value = int(seconds) // period_seconds
            seconds %= period_seconds
            if period_name == "second":
                ms = period_value + total_seconds - int(total_seconds)
                period_str = f"{ms:.2f}".rstrip("0").rstrip(".")
            else:
                period_str = str(period_value)
            texts.append(f"{period_str} {period_name}{'' if period_str == '1' else 's'}")
    return " ".join(texts)



import pytest

from tox.pytest import ToxProject, ToxProjectCreator


@pytest.fixture()
def project(tox_project: ToxProjectCreator) -> ToxProject:
    ini = """
    [tox]
    env_list=py39,py38,py
    [testenv]
    description = with {basepython}
    deps = pypy:
    [testenv:py]
    basepython=py39,py38
    [testenv:fix]
    description = fix it
    """
    return tox_project({"tox.ini": ini})


def test_list_env(project: ToxProject) -> None:
    outcome = project.run("l")

    outcome.assert_success()
    expected = """
    default environments:
    py39 -> with py39
    py38 -> with py38
    py   -> with py39 py38

    additional environments:
    fix  -> fix it
    """
    outcome.assert_out_err(expected, "")


def test_list_env_default(project: ToxProject) -> None:
    outcome = project.run("l", "-d")

    outcome.assert_success()
    expected = """
    default environments:
    py39 -> with py39
    py38 -> with py38
    py   -> with py39 py38
    """
    outcome.assert_out_err(expected, "")


def test_list_env_quiet(project: ToxProject) -> None:
    outcome = project.run("l", "--no-desc")

    outcome.assert_success()
    expected = """
    py39
    py38
    py
    fix
    """
    outcome.assert_out_err(expected, "")


def test_list_env_quiet_default(project: ToxProject) -> None:
    outcome = project.run("l", "--no-desc", "-d")

    outcome.assert_success()
    expected = """
    py39
    py38
    py
    """
    outcome.assert_out_err(expected, "")



from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Type

import pytest

from tox.config.loader.api import Override
from tox.config.loader.memory import MemoryLoader
from tox.config.types import Command, EnvList


def test_memory_loader_repr() -> None:
    loader = MemoryLoader(a=1)
    assert repr(loader) == "MemoryLoader"


def test_memory_loader_override() -> None:
    loader = MemoryLoader(a=1)
    loader.overrides["a"] = Override("a=2")
    loaded = loader.load("a", of_type=int, conf=None, env_name=None)
    assert loaded == 2


@pytest.mark.parametrize(
    ["value", "of_type"],
    [
        (True, bool),
        (1, int),
        ("magic", str),
        ({"1"}, Set[str]),
        ([1], List[int]),
        ({1: 2}, Dict[int, int]),
        (Path.cwd(), Path),
        (Command(["a"]), Command),
        (EnvList("a,b"), EnvList),
        (1, Optional[int]),
        ("1", Optional[str]),
    ],
)
def test_memory_loader(value: Any, of_type: Type[Any]) -> None:
    loader = MemoryLoader(**{"a": value})
    loaded = loader.load("a", of_type=of_type, conf=None, env_name=None)  # noqa
    assert loaded == value


def test_memory_found_keys() -> None:
    loader = MemoryLoader(a=1, c=2)
    assert loader.found_keys() == {"a", "c"}



from argparse import ArgumentTypeError

import pytest
from pytest_mock import MockerFixture

from tox.pytest import MonkeyPatch, ToxProjectCreator
from tox.session.cmd.run.parallel import parse_num_processes
from tox.tox_env.api import ToxEnv
from tox.tox_env.errors import Fail


def test_parse_num_processes_all() -> None:
    assert parse_num_processes("all") is None


def test_parse_num_processes_auto() -> None:
    auto = parse_num_processes("auto")
    assert isinstance(auto, int)
    assert auto > 0


def test_parse_num_processes_exact() -> None:
    assert parse_num_processes("3") == 3


def test_parse_num_processes_not_number() -> None:
    with pytest.raises(ArgumentTypeError, match="value must be a positive number"):
        parse_num_processes("3df")


def test_parse_num_processes_minus_one() -> None:
    with pytest.raises(ArgumentTypeError, match="value must be positive"):
        parse_num_processes("-1")


def test_parallel_general(tox_project: ToxProjectCreator, monkeypatch: MonkeyPatch, mocker: MockerFixture) -> None:
    def setup(self: ToxEnv) -> None:
        if self.conf["env_name"] == "f":
            raise Fail("something bad happened")
        return prev_setup(self)

    prev_setup = ToxEnv.setup
    mocker.patch.object(ToxEnv, "setup", autospec=True, side_effect=setup)
    monkeypatch.setenv("PATH", "")

    ini = """
    [tox]
    no_package=true
    skip_missing_interpreters = true
    env_list= a, b, c, d, e, f
    [testenv]
    commands=python -c 'print("run {env_name}")'
    depends = !c: c
    parallel_show_output = c: true
    [testenv:d]
    base_python = missing_skip
    [testenv:e]
    commands=python -c 'import sys; print("run {env_name}"); sys.exit(1)'
    """
    project = tox_project({"tox.ini": ini})
    outcome = project.run("p", "-p", "all")
    outcome.assert_failed()

    out = outcome.out
    oks, skips, fails = {"a", "b", "c"}, {"d"}, {"e", "f"}
    for env in "a", "b", "c", "d", "e", "f":
        if env in ("c", "e"):
            assert "run c" in out, out
        elif env == "f":
            assert "f: failed with something bad happened" in out, out
        else:
            assert f"run {env}" not in out, out
        of_type = "OK" if env in oks else ("SKIP" if env in skips else "FAIL")
        of_type_icon = "✔" if env in oks else ("⚠" if env in skips else "✖")
        env_done = f"{env}: {of_type} {of_type_icon}"
        assert env_done in out, out

        env_report = f"  {env}: {of_type} {'code 1 ' if env in fails else ''}("
        assert env_report in out, out
        assert out.index(env_done) < out.index(env_report), out


def test_parallel_run_live_out(tox_project: ToxProjectCreator) -> None:
    ini = """
    [tox]
    no_package=true
    env_list= a, b
    [testenv]
    commands=python -c 'print("run {env_name}")'
    """
    project = tox_project({"tox.ini": ini})
    outcome = project.run("p", "-p", "2", "--parallel-live")
    outcome.assert_success()
    assert "python -c" in outcome.out
    assert "run a" in outcome.out
    assert "run b" in outcome.out
